#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Graviti basic file class."""

import mimetypes
from hashlib import sha1
from pathlib import Path
from typing import TYPE_CHECKING, Dict, Optional, Type, TypeVar, Union

import pyarrow as pa
from _io import BufferedReader

from graviti.portex import STANDARD_URL, ExternalElementResgister
from graviti.utility import PathLike, ReprMixin, UserResponse, shorten

if TYPE_CHECKING:
    from graviti.manager import ObjectPermissionManager

_FB = TypeVar("_FB", bound="FileBase")
_F = TypeVar("_F", bound="File")
_RF = TypeVar("_RF", bound="RemoteFile")

_ENCODINGS = mimetypes.encodings_map


class FileBase(ReprMixin):
    """This class represents the file in a DataFrame."""

    __slots__ = ("_key", "_extension", "_size")

    _key: str
    _extension: str
    _size: int

    def _to_post_data(self) -> Dict[str, Union[int, str]]:
        return {"key": self.key, "extension": self.extension, "size": self.size}

    @classmethod
    def _from_pyarrow(
        cls: Type[_FB],
        scalar: pa.StructScalar,
        _: Optional["ObjectPermissionManager"] = None,
    ) -> _FB:
        obj: _FB = object.__new__(cls)
        pyobj = scalar.as_py()

        obj._key = pyobj["key"]
        obj._extension = pyobj["extension"]
        obj._size = pyobj["size"]

        return obj

    @property
    def key(self) -> str:
        """Get the key of the file.

        Returns:
            The key of the file.

        """
        return self._key

    @property
    def extension(self) -> str:
        """Get the extension of the file.

        Returns:
            The extension of the file.

        """
        return self._extension

    @property
    def size(self) -> int:
        """Get the size of the file.

        Returns:
            The size of the file.

        """
        return self._size

    def open(self) -> Union[UserResponse, BufferedReader]:
        """Return the binary file pointer of this file.

        Raises:
            NotImplementedError: The method of the base class should not be called.

        """
        raise NotImplementedError


class File(FileBase):
    """This class represents local files.

    Arguments:
        path: The local path of the file.

    """

    __slots__ = FileBase.__slots__ + ("_path", "_checksum", "_post_key")

    _BUFFER_SIZE = 65536

    _checksum: str
    _post_key: str

    def __init__(self, path: PathLike) -> None:
        self._path = Path(path).absolute()

    def _repr_head(self) -> str:
        return f'{self.__class__.__name__}("{self._path.name}")'

    def _to_post_data(self) -> Dict[str, Union[int, str]]:
        return {"key": self._post_key, "extension": self.extension, "size": self.size}

    @classmethod
    def _from_pyarrow(
        cls: Type[_F],
        scalar: pa.StructScalar,
        _: Optional["ObjectPermissionManager"] = None,
    ) -> _F:
        obj: _F = object.__new__(cls)
        pyobj = scalar.as_py()

        obj._key = pyobj["key"]
        obj._path = Path(pyobj["key"])
        obj._extension = pyobj["extension"]
        obj._size = pyobj["size"]

        return obj

    @property
    def path(self) -> Path:
        """Get the path of the file.

        Returns:
            The path of the file.

        """
        return self._path

    @property
    def key(self) -> str:
        """Get the key of the file.

        Returns:
            The key of the file.

        """
        if not hasattr(self, "_key"):
            self._key = str(self._path)
        return self._key

    @property
    def extension(self) -> str:
        """Get the extension of the file.

        Returns:
            The extension of the file.

        """
        if not hasattr(self, "_extension"):
            suffix = self._path.suffix
            if suffix in _ENCODINGS:
                suffix = Path(self._path.stem).suffix + suffix
            self._extension = suffix
        return self._extension

    @property
    def size(self) -> int:
        """Get the size of the file.

        Returns:
            The size of the file.

        """
        if not hasattr(self, "_size"):
            self._size = self._path.stat().st_size
        return self._size

    def get_checksum(self) -> str:
        """Get the sha1 checksum of the local file.

        Returns:
            The sha1 checksum of the local file.

        """
        if not hasattr(self, "_checksum"):
            sha1_object = sha1()
            with self._path.open("rb") as fp:
                while True:
                    data = fp.read(self._BUFFER_SIZE)
                    if not data:
                        break
                    sha1_object.update(data)

            self._checksum = sha1_object.hexdigest()

        return self._checksum

    def open(self) -> BufferedReader:
        """Return the binary file pointer of this file.

        Returns:
            The local file pointer.

        """
        return self._path.open("rb")


@ExternalElementResgister(STANDARD_URL, "main", "file.File")
class RemoteFile(FileBase):
    """This class represents the file on Graviti platform.

    Arguments:
        key: The key of the file.
        extension: The extension of the file.
        size: The size of the file.
        object_permission_manager: The permission to access the file.

    """

    __slots__ = FileBase.__slots__ + ("_object_permission",)

    def __init__(
        self,
        key: str,
        extension: str,
        size: int,
        object_permission_manager: "ObjectPermissionManager",
    ) -> None:
        self._key = key
        self._extension = extension
        self._size = size
        self._object_permission = object_permission_manager

    def _repr_head(self) -> str:
        short_checksum = shorten(self.key.rsplit("/", 1)[1])
        return f'{self.__class__.__name__}("{short_checksum}")'

    @classmethod
    def _from_pyarrow(  # type: ignore[override]  # pylint: disable=signature-differs
        cls: Type[_RF],
        scalar: pa.StructScalar,
        object_permission_manager: "ObjectPermissionManager",
    ) -> _RF:
        obj: _RF = object.__new__(cls)
        pyobj = scalar.as_py()

        obj._key = pyobj["key"]
        obj._extension = pyobj["extension"]
        obj._size = pyobj["size"]
        obj._object_permission = object_permission_manager

        return obj

    def open(self) -> UserResponse:
        """Return the binary file pointer of this file.

        Returns:
            The remote file pointer.

        """
        return self._object_permission.get_object(self._key)



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Graviti point cloud file class."""

from graviti.file.base import File, RemoteFile
from graviti.portex import STANDARD_URL, ExternalElementResgister


class PointCloud(File):
    """This class represents local point cloud files."""

    __slots__ = File.__slots__


@ExternalElementResgister(
    STANDARD_URL,
    "main",
    "file.PointCloud",
    "file.PointCloudBin",
)
class RemotePointCloud(RemoteFile):
    """This class represents remote point cloud files."""

    __slots__ = RemoteFile.__slots__



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Interfaces about the draft."""

from typing import Any, Dict, Optional

from graviti.openapi.requests import open_api_do


def create_draft(
    access_key: str,
    url: str,
    owner: str,
    dataset: str,
    *,
    title: str,
    branch: Optional[str] = None,
    description: Optional[str] = None,
) -> Dict[str, Any]:
    """Execute the OpenAPI `POST /v2/datasets/{owner}/{dataset}/drafts`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        owner: The owner of the dataset.
        dataset: Name of the dataset, unique for a user.
        title: The draft title.
        branch: The specified branch name. None means the default branch of the dataset.
        description: The draft description.

    Returns:
        The response of OpenAPI.

    Examples:
        >>> create_draft(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "graviti-example",
        ...     "MNIST",
        ...     title="draft-2",
        ...     branch="main",
        ... )
        {
           "number": 2,
           "title": "draft-2",
           "description": "",
           "branch": "main",
           "state": "OPEN",
           "parent_commit_id": "85c57a7f03804ccc906632248dc8c359",
           "creator": "graviti-example",
           "created_at": "2021-03-03T18:58:10Z",
           "updated_at": "2021-03-03T18:58:10Z"
        }

    """
    url = f"{url}/v2/datasets/{owner}/{dataset}/drafts"
    post_data = {"title": title}

    if description:
        post_data["description"] = description
    if branch:
        post_data["branch"] = branch

    return open_api_do(  # type: ignore[no-any-return]
        "POST", access_key, url, json=post_data
    ).json()


def list_drafts(
    access_key: str,
    url: str,
    owner: str,
    dataset: str,
    *,
    state: Optional[str] = None,
    branch: Optional[str] = None,
    offset: Optional[int] = None,
    limit: Optional[int] = None,
) -> Dict[str, Any]:
    """Execute the OpenAPI `GET /v2/datasets/{owner}/{dataset}/drafts`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        owner: The owner of the dataset.
        dataset: Name of the dataset, unique for a user.
        state: The draft state which includes "OPEN", "CLOSED", "COMMITTED", "ALL" and None.
            None means listing open drafts.
        branch: The branch name. None means listing drafts from all branches.
        offset: The offset of the page. The default value of this param in OpenAPIv2 is 0.
        limit: The limit of the page. The default value of this param in OpenAPIv2 is 128.

    Returns:
        The response of OpenAPI.

    Examples:
        >>> list_drafts(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "graviti-example",
        ...     "MNIST",
        ... )
        {
           "drafts": [
               {
                   "number": 2,
                   "title": "draft-2",
                   "description": "",
                   "branch": "main",
                   "state": "OPEN",
                   "parent_commit_id": "85c57a7f03804ccc906632248dc8c359",
                   "creator": "graviti-example",
                   "created_at": "2021-03-03T18:58:10Z",
                   "updated_at": "2021-03-03T18:58:10Z"
               }
           ],
           "offset": 0,
           "record_size": 1,
           "total_count": 1
        }

    """
    url = f"{url}/v2/datasets/{owner}/{dataset}/drafts"

    params: Dict[str, Any] = {}
    if state:
        params["state"] = state
    if branch:
        params["branch"] = branch
    if offset is not None:
        params["offset"] = offset
    if limit is not None:
        params["limit"] = limit

    return open_api_do("GET", access_key, url, params=params).json()  # type: ignore[no-any-return]


def get_draft(
    access_key: str, url: str, owner: str, dataset: str, *, draft_number: int
) -> Dict[str, Any]:
    """Execute the OpenAPI `GET /v2/datasets/{owner}/{dataset}/drafts/{draft_number}`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        owner: The owner of the dataset.
        dataset: Name of the dataset, unique for a user.
        draft_number: Number of the draft.

    Returns:
        The response of OpenAPI.

    Examples:
        >>> get_draft(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "MNIST",
        ...     "graviti-example",
        ...     draft_number=2,
        ... )
        {
           "number": 2,
           "title": "draft-2",
           "description": "",
           "branch": "main",
           "state": "OPEN",
           "parent_commit_id": "85c57a7f03804ccc906632248dc8c359",
           "creator": "graviti-example",
           "created_at": "2021-03-03T18:58:10Z",
           "updated_at": "2021-03-03T18:58:10Z"
        }

    """
    url = f"{url}/v2/datasets/{owner}/{dataset}/drafts/{draft_number}"
    return open_api_do("GET", access_key, url).json()  # type: ignore[no-any-return]


def update_draft(
    access_key: str,
    url: str,
    owner: str,
    dataset: str,
    *,
    draft_number: int,
    state: Optional[str] = None,
    title: Optional[str] = None,
    description: Optional[str] = None,
) -> Dict[str, Any]:
    """Execute the OpenAPI `PATCH /v2/datasets/{owner}/{dataset}/drafts/{draft_number}`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        owner: The owner of the dataset.
        dataset: Name of the dataset, unique for a user.
        draft_number: The updated draft number.
        state: The updated draft state which could be "CLOSED" or None.
            Where None means no change in state.
        title: The draft title.
        description: The draft description.

    Returns:
        The response of OpenAPI.

    Examples:
        Update the title or description of the draft:

        >>> update_draft(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "MNIST",
        ...     draft_number=2,
        ...     title="draft-3"
        ... )
        {
           "number": 2,
           "title": "draft-3",
           "description": "",
           "branch": "main",
           "state": "OPEN",
           "parent_commit_id": "85c57a7f03804ccc906632248dc8c359",
           "creator": "graviti-example",
           "created_at": "2021-03-03T18:58:10Z",
           "updated_at": "2021-03-04T18:58:10Z"
        }

        Close the draft:

        >>> update_draft(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "MNIST",
        ...     draft_number=2,
        ...     state="CLOSED"
        ... )
        {
           "number": 2,
           "title": "draft-3",
           "description": "",
           "branch": "main",
           "state": "CLOSED",
           "parent_commit_id": "85c57a7f03804ccc906632248dc8c359",
           "creator": "graviti-example",
           "created_at": "2021-03-03T18:58:10Z",
           "updated_at": "2021-03-05T18:58:10Z"
        }

    """
    url = f"{url}/v2/datasets/{owner}/{dataset}/drafts/{draft_number}"
    patch_data: Dict[str, Any] = {"draft_number": draft_number}

    if state:
        patch_data["state"] = state
    if title is not None:
        patch_data["title"] = title
    if description is not None:
        patch_data["description"] = description

    return open_api_do(  # type: ignore[no-any-return]
        "PATCH", access_key, url, json=patch_data
    ).json()



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#
"""Parameter releated classes."""


from collections import OrderedDict
from copy import deepcopy
from inspect import Parameter, Signature
from typing import Any, Dict, Iterable, List, Mapping, MutableMapping, Optional

from graviti.portex import ptype as PTYPE
from graviti.portex.package import Imports
from graviti.utility import UserMapping


def param(
    default: Any = Parameter.empty,
    options: Optional[Iterable[Any]] = None,
    ptype: PTYPE.PType = PTYPE.Any,
) -> Any:
    """The factory function of Param.

    Arguments:
        default: The default value of the parameter.
        options: All possible values of the parameter.
        ptype: The parameter type.

    Returns:
        A tuple which contains "default", "options" and "ptype".

    """
    return default, options, ptype


class Param(Parameter):
    """Represents a parameter of a portex type.

    Arguments:
        name: The name of the parameter.
        default: The default value of the parameter.
        options: All possible values of the parameter.
        ptype: The parameter type.

    """

    _empty = Parameter.empty

    def __init__(
        self,
        name: str,
        default: Any = _empty,
        options: Optional[Iterable[Any]] = None,
        ptype: PTYPE.PType = PTYPE.Any,
    ) -> None:
        if default is not self._empty:
            default = ptype.load(default)
        super().__init__(name, Parameter.POSITIONAL_OR_KEYWORD, default=default)
        self.options = set(options) if options else None
        self.ptype = ptype

    @classmethod
    def from_pyobj(cls, pyobj: Dict[str, Any], ptype: PTYPE.PType = PTYPE.Any) -> "Param":
        """Create Param instance from python dict.

        Arguments:
            pyobj: A python dict representing a parameter.
            ptype: The parameter type.

        Returns:
            A Param instance created from the input python dict.

        """
        return cls(
            pyobj["name"],
            pyobj.get("default", cls._empty),
            pyobj.get("options"),
            ptype,
        )

    @property
    def required(self) -> bool:
        """Whether this parameter is a required parameter.

        Returns:
            ``True`` for required and ``False`` for non-required parameter.

        """
        return self.default == self._empty  # type: ignore[no-any-return]

    def to_pyobj(self) -> Dict[str, Any]:
        """Dump the instance to a python dict.

        Returns:
            A python dict representation of the Param.

        """
        pyobj: Dict[str, Any] = {"name": self.name}
        if self.default != self._empty:
            pyobj["default"] = self.default
        if self.options is not None:
            pyobj["options"] = list(self.options)

        return pyobj

    def check(self, arg: Any) -> Any:
        """Check the validity of the parameter.

        Arguments:
            arg: The argument which needs to be checked.

        Returns:
            The argument after checking.

        Raises:
            ValueError: Raise when the argument is not in options.

        """
        if arg is self.default:
            return deepcopy(arg)

        arg = self.ptype.check(arg)
        # https://github.com/PyCQA/pylint/issues/3045
        # pylint: disable=unsupported-membership-test
        if self.options and arg not in self.options:
            raise ValueError(f"Arguments '{arg}' is not in {self.options}")

        return arg

    def load(self, content: Any, imports: Imports) -> Any:
        """Create an instance of the parameter type from the python content.

        Arguments:
            content: A python presentation of the parameter type.
            imports: The imports of the parameter type.

        Returns:
            An instance of the parameter type.

        """
        return self.ptype.load(content, imports)

    def dump(self, arg: Any) -> Any:
        """Dump the parameter type instance into the python presentation.

        Arguments:
            arg: The parameter type instance.

        Returns:
            The python presentation of the input instance.

        """
        return self.ptype.dump(arg)


class Params(UserMapping[str, Param]):
    """Represents all parameters of a portex type.

    Arguments:
        values: The parameters mapping.

    """

    def __init__(self, values: Optional[Mapping[str, Param]] = None) -> None:
        # Why not use typing.OrderedDict here?
        # typing.OrderedDict is supported after python 3.7.2
        # typing_extensions.OrderedDict will trigger https://github.com/python/mypy/issues/11528
        self._data: MutableMapping[str, Param] = OrderedDict(values if values else {})

    @classmethod
    def from_pyobj(cls, pyobj: List[Dict[str, Any]], keys: Dict[str, Any]) -> "Params":
        """Create Params instance from python list.

        Arguments:
            pyobj: A python dict representing parameters.
            keys: A python dict containing parameter types.

        Returns:
            A Params instance created from the input python list.

        """
        params = cls()
        for item in pyobj:
            params.add(Param.from_pyobj(item, keys.get(item["name"], PTYPE.Any)))

        return params

    def to_pyobj(self) -> List[Dict[str, Any]]:
        """Dump the instance to a python list.

        Returns:
            A python list representation of the Params.

        """
        return [item.to_pyobj() for item in self._data.values()]

    def add(self, value: Param) -> None:
        """Add a parameter.

        Arguments:
            value: The parameter which needs to be added to this instance.

        Raises:
            KeyError: When the parameter name is duplicated.

        """
        name = value.name
        if name in self._data:
            raise KeyError(f"Duplicate parameter name: '{name}'")

        self._data[name] = value

    def update(self, values: Mapping[str, Param]) -> None:
        """Update the parameters.

        Arguments:
            values: The parameters which need to be updated to this instance.

        """
        for value in values.values():
            self.add(value)

    def get_signature(self) -> Signature:
        """Get the python inspect Signature from parameters.

        Returns:
            The :class:`Signature` instance created by all parameters in this instance.

        """
        return Signature(list(self._data.values()))



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Graviti Python SDK."""

from graviti.__version__ import __version__
from graviti.dataframe import ColumnSeries as Series
from graviti.dataframe import DataFrame
from graviti.utility import engine
from graviti.workspace import Workspace

__all__ = ["DataFrame", "Series", "Workspace", "__version__", "engine"]



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Related classes for the lazy evaluation."""

from itertools import repeat, zip_longest
from typing import (
    Any,
    Callable,
    Generator,
    Generic,
    Iterable,
    Iterator,
    List,
    MutableSequence,
    Optional,
    Tuple,
    TypeVar,
    Union,
    overload,
)

from graviti.utility import ReprMixin, ReprType, locked

_T = TypeVar("_T")
PagingGenerator = Callable[[int, int], Generator[_T, None, int]]


class LazyItem(Generic[_T]):
    """In paging lazy evaluation system, a LazyItem instance represents an element in a pagination.

    If user wants to access the elememt, LazyItem will trigger the paging request to pull a page of
    elements and return the required element. All the pulled elements will be stored in different
    LazyItem instances and will not be requested again.

    Arguments:
        page: The page the item belongs to.

    Attributes:
        page: The parent :class:`LazyPage` of this item.
        data: The actual element stored in this item.

    """

    __slots__ = ("page", "data")

    def __init__(self, page: "LazyPage[_T]", data: _T):
        self.page = page
        self.data = data

    @classmethod
    def from_page(cls, page: "LazyPage[_T]") -> "LazyItem[_T]":
        """Create a LazyItem instance from page.

        Arguments:
            page: The page of the element.

        Returns:
            The LazyItem instance which stores the input page.

        """
        obj: "LazyItem[_T]" = object.__new__(cls)
        obj.page = page
        return obj

    @classmethod
    def from_data(cls, data: _T) -> "LazyItem[_T]":
        """Create a LazyItem instance from data.

        Arguments:
            data: The actual data needs to be stored in LazyItem.

        Returns:
            The LazyItem instance which stores the input data.

        """
        obj: "LazyItem[_T]" = object.__new__(cls)
        obj.data = data
        return obj

    def get(self) -> _T:
        """Access the actual element represented by LazyItem.

        If the element is already pulled from web, it will be return directly, otherwise this
        function will request for a page of elements to get the required elememt.

        Returns:
            The actual element this LazyItem instance represents.

        """
        if not hasattr(self, "data"):
            self.page.pull()

        return self.data


_R = TypeVar("_R")


class ReturnGenerator(Generic[_T, _R]):
    """ReturnGenerator is a generator wrap to get the return value easily.

    Arguments:
        generator: The generator needs to be wrapped.

    Attributes:
        value: The return value of the input generator.

    """

    value: _R

    def __init__(self, generator: Generator[_T, Any, _R]):
        self._generator = generator

    def __iter__(self) -> Iterator[_T]:
        self.value = yield from self._generator


class LazyPage(Generic[_T]):
    """In paging lazy evaluation system, a LazyPage instance represents a page with elements.

    LazyPage is used for sending paging request to pull a page of elements and storing them in
    different :class:`LazyItem` instances.

    Arguments:
        offset: The offset of the page.
        limit: The limit of the page.
        func: A paging generator function, which takes offset<int> and limit<int> as inputs and
            returns a generator. The returned generator should yield the element user needs, and
            return the total count of the elements in the paging request.

    Attributes:
        items: The :class:`LazyItem` list which represents a page of elements.

    """

    __slots__: Tuple[str, ...] = ("_offset", "_limit", "_func", "items")

    def __init__(self, offset: int, limit: int, func: PagingGenerator[_T]) -> None:
        self.items: Tuple[LazyItem[_T], ...] = tuple(LazyItem.from_page(self) for _ in range(limit))

        self._init(offset, limit, func)

    def _init(self, offset: int, limit: int, func: PagingGenerator[_T]) -> None:
        self._offset = offset
        self._limit = limit
        self._func = func

    @classmethod
    def from_items(
        cls,
        offset: int,
        limit: int,
        func: PagingGenerator[_T],
        item_contents: Iterable[_T],
    ) -> "LazyPage[_T]":
        """Create a LazyPage instance with the given items and generator function.

        Arguments:
            offset: The offset of the page.
            limit: The limit of the page.
            func: A paging generator function, which takes offset<int> and limit<int> as inputs and
                returns a generator. The returned generator should yield the element user needs, and
                return the total count of the elements in the paging request.
            item_contents: The lazy item contents that need to be stored on this page.

        Returns:
            The LazyPage instance which stores the input items and function.

        """
        obj: "LazyPage[_T]" = object.__new__(cls)
        obj._init(offset, limit, func)
        obj.items = tuple(LazyItem(obj, item) for item in item_contents)

        return obj

    @locked
    def pull(self) -> None:
        """Send paging request to pull a page of elements and store them in :class:`LazyItem`."""
        for data, item in zip(self._func(self._offset, self._limit), self.items):
            item.data = data


class InitPage(LazyPage[_T]):
    """In paging lazy evaluation system, InitPage is the page to initialize :class:`LazyPagingList`.

    InitPage will send a paging request to pull a page of elements and storing them in different
    :class:`LazyItem` instances when construction. And the totalCount of the page will also be
    stored in the instance.

    Arguments:
        offset: The offset of the page.
        limit: The limit of the page.
        func: A paging generator function, which takes offset<int> and limit<int> as inputs and
            returns a generator. The returned generator should yield the element user needs, and
            return the total count of the elements in the paging request.

    Attributes:
        items: The :class:`LazyItem` list which represents a page of elements.
        total_count: The totalCount of the paging request.

    """

    __slots__ = LazyPage.__slots__ + ("total_count",)

    def __init__(  # pylint: disable=super-init-not-called
        self, offset: int, limit: int, func: PagingGenerator[_T]
    ) -> None:
        generator = ReturnGenerator(func(offset, limit))
        self.items: Tuple[LazyItem[_T], ...] = tuple(LazyItem(self, data) for data in generator)

        self._init(offset, len(self.items), func)

        self.total_count = generator.value


class LazyPagingList(MutableSequence[_T], ReprMixin):
    """LazyPagingList is a wrap of web paging request.

    It follows the python MutableSequence protocal, which means it can be used like a python builtin
    list. And it provides features like lazy evaluation and cache.

    Arguments:
        func: A paging generator function, which takes offset<int> and limit<int> as inputs and
            returns a generator. The returned generator should yield the element user needs, and
            return the total count of the elements in the paging request.
        limit: The page size of each paging request.

    """

    _S = TypeVar("_S", bound="LazyPagingList[_T]")

    _repr_type = ReprType.SEQUENCE

    _items: List[LazyItem[_T]]

    def __init__(self, func: PagingGenerator[_T], limit: int) -> None:
        self._func = func
        self._limit = limit
        self._init_items: Callable[[int], None] = self._init_all_items

    def __len__(self) -> int:
        return self._get_items().__len__()

    @overload
    def __getitem__(self, index: int) -> _T:
        ...

    @overload
    def __getitem__(self: _S, index: slice) -> _S:
        ...

    def __getitem__(self: _S, index: Union[int, slice]) -> Union[_T, _S]:
        if isinstance(index, slice):
            return self._get_slice(index)

        return self._get_items(index)[index].get()

    @overload
    def __setitem__(self, index: int, value: _T) -> None:
        ...

    @overload
    def __setitem__(self, index: slice, value: Iterable[_T]) -> None:
        ...

    def __setitem__(self, index: Union[int, slice], value: Union[_T, Iterable[_T]]) -> None:
        # https://github.com/python/mypy/issues/7858
        if isinstance(index, slice):
            self._get_items().__setitem__(
                index,
                map(LazyItem.from_data, value),  # type: ignore[arg-type]
            )
            return

        self._get_items(index).__setitem__(
            index,
            LazyItem.from_data(value),  # type: ignore[arg-type]
        )

    def __delitem__(self, index: Union[int, slice]) -> None:
        self._get_items().__delitem__(index)

    def __iter__(self) -> Iterator[_T]:
        for item in self._get_items():
            yield item.get()

    def __reversed__(self) -> Iterator[_T]:
        for item in self._get_items().__reversed__():
            yield item.get()

    def __contains__(self, value: Any) -> bool:
        for item in self._get_items():
            if item.get() == value:
                return True

        return False

    def __iadd__(self: _S, values: Iterable[_T]) -> _S:
        self._get_items().__iadd__(LazyItem.from_data(value) for value in values)
        return self

    @staticmethod
    def _range(total_count: int, limit: int) -> Iterator[Tuple[int, int]]:
        """A Generator which generates offset and limit for paging request.

        Examples:
            >>> self._range(10, 3)
            <generator object paging_range at 0x11b9932e0>

            >>> list(self._range(10, 3))
            [(0, 3), (3, 3), (6, 3), (9, 1)]

        Arguments:
            total_count: The total count of the page.
            limit: The paging limit.

        Yields:
            The tuple (offset, limit) for paging request.

        """
        div, mod = divmod(total_count, limit)
        yield from zip_longest(range(0, total_count, limit), repeat(limit, div), fillvalue=mod)

    @locked
    def _init_all_items(self, index: int = 0) -> None:
        index = index if index >= 0 else 0
        index_offset = index // self._limit * self._limit
        init_page = InitPage(index_offset, self._limit, self._func)
        total_count = init_page.total_count
        self._items: List[LazyItem[_T]] = []
        for offset, limit in self._range(total_count, self._limit):
            page = init_page if offset == index_offset else LazyPage(offset, limit, self._func)
            self._items.extend(page.items)

    @locked
    def _init_sliced_items(self: _S, parent: _S, slicing: slice) -> None:
        self._items = parent._get_items()[slicing]  # pylint: disable=protected-access

    def _get_items(self, index: int = 0) -> List[LazyItem[_T]]:
        if not hasattr(self, "_items"):
            self._init_items(index)

        return self._items

    def _get_slice(self: _S, slicing: slice) -> _S:
        # pylint: disable=protected-access
        paging_list = self.__class__(self._func, self._limit)
        if hasattr(self, "_items"):
            paging_list._items = self._items[slicing]
        else:
            paging_list._init_items = lambda _: paging_list._init_sliced_items(self, slicing)

        return paging_list

    def insert(self, index: int, value: _T) -> None:
        """Insert object before index.

        Arguments:
            index: Position of the LazyPagingList.
            value: Element to be inserted into the LazyPagingList.

        """
        self._get_items(index).insert(index, LazyItem.from_data(value))

    def append(self, value: _T) -> None:
        """Append object to the end of the LazyPagingList.

        Arguments:
            value: Element to be appended to the LazyPagingList.

        """
        self._get_items().append(LazyItem.from_data(value))

    def reverse(self) -> None:
        """Reverse the items of the LazyPagingList in place."""
        self._get_items().reverse()

    def pop(self, index: int = -1) -> _T:
        """Return the item at index (default last) and remove it from the LazyPagingList.

        Arguments:
            index: Position of the LazyPagingList.

        Returns:
            Element to be removed from the LazyPagingList.

        """
        return self._get_items(index).pop(index).get()

    def index(self, value: Any, start: int = 0, stop: Optional[int] = None) -> int:
        """Return the first index of the value.

        Arguments:
            value: The value to be found.
            start: The start index of the subsequence.
            stop: The end index of the subsequence.

        Raises:
            ValueError: When the value is not in the LazyPagingList

        Returns:
            The first index of the value.


        """
        items = self._get_items(start)
        length = len(items)

        stop = length if stop is None else min(stop, length)

        for i in range(start, stop):
            if items[i].get() == value:
                return i

        raise ValueError(f"{value} is not in LazyPagingList")

    def count(self, value: Any) -> int:
        """Return the number of occurrences of value.

        Arguments:
            value: The value needs to be counted.

        Returns:
            The number of occurrences of value.

        """
        return sum(1 for item in self._get_items() if item.get() == value)

    def extend(self, values: Iterable[_T]) -> None:
        """Extend LazyPagingList by appending elements from the iterable.

        Arguments:
            values: Elements to be extended into the LazyPagingList.

        """
        self._get_items().extend(LazyItem.from_data(value) for value in values)



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Engine control related classes."""

from enum import Enum, auto
from typing import Any


class Mode(Enum):
    """This class defines the engine mode and includes 'LOCAl' and 'ONLINE'."""

    LOCAL = auto()
    ONLINE = auto()


class Online:
    """An engine controller used to start and stop the online mode."""

    _old_mode: Mode

    def __init__(self, _engine: "Engine") -> None:
        self._engine = _engine

    def __enter__(self) -> None:
        self._old_mode = self._engine.mode
        self._engine.mode = Mode.ONLINE

    def __exit__(self, *_: Any) -> None:
        self._engine.mode = self._old_mode


class Engine:
    """This is a base class defining the Engine mode."""

    mode = Mode.LOCAL

    def online(self) -> Online:
        """Init a Online instance.

        Returns:
            the Online instance.

        """
        return Online(self)


engine = Engine()



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Graviti customized types."""

import inspect
from pathlib import Path
from typing import AbstractSet, Any, Tuple, Type, TypeVar, Union

from typing_extensions import Protocol

_K = TypeVar("_K")
_V = TypeVar("_V")

PathLike = Union[str, Path]


class NestedDict(Protocol[_K, _V]):
    """Typehint for nested dict."""

    def __contains__(self, key: _K) -> bool:
        ...

    def __getitem__(self, key: _K) -> Union["NestedDict[_K, _V]", _V]:
        ...

    def __setitem__(self, key: _K, value: Union["NestedDict[_K, _V]", _V]) -> None:
        ...

    def items(self) -> AbstractSet[Tuple[_K, Union["NestedDict[_K, _V]", _V]]]:
        """Return (key, value) pairs of the dict."""
        ...

    def setdefault(
        self, key: _K, default: Union["NestedDict[_K, _V]", _V]
    ) -> Union["NestedDict[_K, _V]", _V]:
        """Get the value of the key if exists, else set the value as default and return.

        Arguments:
            key: The key.
            default: The default value to set if the key does not exist.

        """
        ...


def check_type(name: str, value: Any, expected_type: Type[Any]) -> None:
    """Check the type of the argument.

    Arguments:
        name: The name of the argument.
        value: The value of the argument.
        expected_type: The type of the argument.

    Raises:
        TypeError: When the value is not of the type.

    """
    if not isinstance(value, expected_type):
        raise TypeError(
            f'{inspect.stack()[1][3]}(): "{name}" must be {expected_type}, not {type(value)}'
        )



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Paging list related class."""

from itertools import repeat
from math import ceil
from typing import Any, Callable, Iterator, List, Optional, Tuple, TypeVar

import pyarrow as pa

from graviti.paging.lists import MappedPagingList, PagingList, PyArrowPagingList
from graviti.paging.offset import Offsets
from graviti.paging.wrapper import StructArrayWrapper

_T = TypeVar("_T")


class LazyFactoryBase:
    """LazyFactoryBase is the base class of the lazy facotry."""

    _patype: pa.DataType

    def __getitem__(self, key: str) -> "LazySubFactory":
        raise NotImplementedError

    def __contains__(self, key: str) -> bool:
        try:
            self._patype.__getitem__(key)
            return True
        except KeyError:
            return False

    def create_list(self, mapper: Callable[[Any], _T]) -> PagingList[_T]:
        """Create a paging list from the factory.

        Arguments:
            mapper: A callable object to convert every item in the pyarrow array.

        Raises:
            NotImplementedError: The method of the base class should not be called.

        """
        raise NotImplementedError

    def create_mapped_list(self, mapper: Callable[[Any], _T]) -> MappedPagingList[_T]:
        """Create a paging list from the factory.

        Arguments:
            mapper: A callable object to convert every item in the pyarrow array.

        Raises:
            NotImplementedError: The method of the base class should not be called.

        """
        raise NotImplementedError

    def create_pyarrow_list(self) -> PyArrowPagingList[Any]:
        """Create a paging list from the factory.

        Raises:
            NotImplementedError: The method of the base class should not be called.

        """
        raise NotImplementedError


class LazyFactory(LazyFactoryBase):
    """LazyFactory is a factory for requesting source data and creating paging lists.

    Arguments:
        total_count: The total count of the elements in the paging lists.
        limit: The size of each lazy load page.
        getter: A callable object to get the source data.
        patype: The pyarrow DataType of the data in the factory.

    Examples:
        >>> import pyarrow as pa
        >>> patype = pa.struct(
        ...     {
        ...         "remotePath": pa.string(),
        ...         "label": pa.struct({"CLASSIFICATION": pa.struct({"category": pa.string()})}),
        ...     }
        ... )
        >>> TOTAL_COUNT = 1000
        >>> def getter(offset: int, limit: int) -> List[Dict[str, Any]]:
        ...     stop = min(offset + limit, TOTAL_COUNT)
        ...     return [
        ...         {
        ...             "remotePath": f"{i:06}.jpg",
        ...             "label": {"CLASSIFICATION": {"category": "cat" if i % 2 else "dog"}},
        ...         }
        ...         for i in range(offset, stop)
        ...     ]
        ...
        >>> factory = LazyFactory(TOTAL_COUNT, 128, getter, patype)
        >>> paths = factory["remotePath"].create_pyarrow_list()
        >>> categories = factory["label"]["CLASSIFICATION"]["category"].create_pyarrow_list()
        >>> len(paths)
        1000
        >>> list(paths)
        [<pyarrow.StringScalar: '000000.jpg'>,
         <pyarrow.StringScalar: '000001.jpg'>,
         <pyarrow.StringScalar: '000002.jpg'>,
         <pyarrow.StringScalar: '000003.jpg'>,
         <pyarrow.StringScalar: '000004.jpg'>,
         <pyarrow.StringScalar: '000005.jpg'>,
         ...
         ...
         <pyarrow.StringScalar: '000999.jpg'>]
        >>> len(categories)
        1000
        >>> list(categories)
        [<pyarrow.StringScalar: 'dog'>,
         <pyarrow.StringScalar: 'cat'>,
         <pyarrow.StringScalar: 'dog'>,
         <pyarrow.StringScalar: 'cat'>,
         <pyarrow.StringScalar: 'dog'>,
         ...
         ...
         <pyarrow.StringScalar: 'cat'>]


    """

    def __init__(
        self, total_count: int, limit: int, getter: Callable[[int, int], Any], patype: pa.DataType
    ) -> None:
        self._getter = getter
        self._total_count = total_count
        self._limit = limit
        self._patype = patype
        self._pages: List[Optional[pa.StructArray]] = [None] * ceil(total_count / limit)

    def __getitem__(self, key: str) -> "LazySubFactory":
        return LazySubFactory(self, (key,), self._patype[key].type)

    def get_array(self, pos: int, keys: Tuple[str, ...]) -> pa.Array:
        """Get the array from the factory.

        Arguments:
            pos: The page number.
            keys: The keys to access the array from factory.

        Returns:
            The requested pyarrow array.

        """
        array = self._pages[pos]
        if array is None:
            array = pa.array(self._getter(pos * self._limit, self._limit), type=self._patype)
            self._pages[pos] = array

        for key in keys:
            array = array.field(key)

        return array

    def create_list(self, mapper: Callable[[Any], _T]) -> PagingList[_T]:
        """Create a paging list from the factory.

        Arguments:
            mapper: A callable object to convert every item in the pyarrow array.

        Returns:
            A paging list created from the factory.

        """
        return PagingList.from_factory(self, (), mapper)

    def create_mapped_list(self, mapper: Callable[[Any], _T]) -> MappedPagingList[_T]:
        """Create a paging list from the factory.

        Arguments:
            mapper: A callable object to convert every item in the pyarrow array.

        Returns:
            A paging list created from the factory.

        """
        return MappedPagingList.from_factory(self, (), mapper)

    def create_pyarrow_list(self) -> PyArrowPagingList[Any]:
        """Create a paging list from the factory.

        Returns:
            A paging list created from the factory.

        """
        return PyArrowPagingList.from_factory(self, (), self._patype)

    def get_page_lengths(self) -> Iterator[int]:
        """A Generator which generates the length of the pages in the factory.

        Yields:
            The page lengths.

        """
        div, mod = divmod(self._total_count, self._limit)
        yield from repeat(self._limit, div)
        if mod != 0:
            yield mod

    def get_offsets(self) -> Offsets:
        """Get the Offsets instance created by the total_count and limit of this factory.

        Returns:
            The Offsets instance created by the total_count and limit of this factory.

        """
        return Offsets(self._total_count, self._limit)


class LazySubFactory(LazyFactoryBase):
    """LazySubFactory is a factory for creating paging lists.

    Arguments:
        factory: The source LazyFactory instance.
        keys: The keys to access the array from the source LazyFactory.
        patype: The pyarrow DataType of the data in the sub-factory.

    """

    def __init__(self, factory: LazyFactory, keys: Tuple[str, ...], patype: pa.DataType) -> None:
        self._factory = factory
        self._keys = keys
        self._patype = patype

    def __getitem__(self, key: str) -> "LazySubFactory":
        return LazySubFactory(self._factory, self._keys + (key,), self._patype[key].type)

    def create_list(self, mapper: Callable[[Any], _T]) -> PagingList[_T]:
        """Create a paging list from the factory.

        Arguments:
            mapper: A callable object to convert every item in the pyarrow array.

        Returns:
            A paging list created from the factory.

        """
        return PagingList.from_factory(self._factory, self._keys, mapper)

    def create_mapped_list(self, mapper: Callable[[Any], _T]) -> MappedPagingList[_T]:
        """Create a paging list from the factory.

        Arguments:
            mapper: A callable object to convert every item in the pyarrow array.

        Returns:
            A paging list created from the factory.

        """
        return MappedPagingList.from_factory(self._factory, self._keys, mapper)

    def create_pyarrow_list(self) -> PyArrowPagingList[Any]:
        """Create a paging list from the factory.

        Returns:
            A paging list created from the factory.

        """
        return PyArrowPagingList.from_factory(self._factory, self._keys, self._patype)


class LazyLowerCaseFactory(LazyFactory):
    """LazyLowerCaseFactory is a factory to handle the case insensitive data from graviti back-end.

    Arguments:
        total_count: The total count of the elements in the paging lists.
        limit: The size of each lazy load page.
        getter: A callable object to get the source data.
        patype: The pyarrow DataType of the data in the factory.

    """

    def __init__(
        self, total_count: int, limit: int, getter: Callable[[int, int], Any], patype: pa.DataType
    ) -> None:
        super().__init__(total_count, limit, getter, self._lower_patype(patype))

    def __getitem__(self, key: str) -> "LazyLowerCaseSubFactory":
        lower_key = key.lower()
        return LazyLowerCaseSubFactory(self, (lower_key,), self._patype[lower_key].type)

    def _lower_patype(self, patype: pa.DataType) -> pa.DataType:
        if isinstance(patype, pa.StructType):
            return pa.struct(
                {field.name.lower(): self._lower_patype(field.type) for field in patype}
            )

        if isinstance(patype, pa.ListType):
            return pa.list_(self._lower_patype(patype.value_type))

        return patype

    def get_array(self, pos: int, keys: Tuple[str, ...]) -> pa.Array:
        """Get the array from the factory.

        Arguments:
            pos: The page number.
            keys: The keys to access the array from factory.

        Returns:
            The requested pyarrow array.

        """
        array = self._pages[pos]
        if array is None:
            array = pa.array(self._getter(pos * self._limit, self._limit), type=self._patype)
            array = StructArrayWrapper(array)
            self._pages[pos] = array

        for key in keys:
            array = array.field(key)

        return array


class LazyLowerCaseSubFactory(LazySubFactory):
    """LazyLowerCaseSubFactory is a sub-factory to handle the case insensitive data.

    Arguments:
        factory: The source LazyFactory instance.
        keys: The keys to access the array from the source LazyFactory.
        patype: The pyarrow DataType of the data in the sub-factory.

    """

    def __getitem__(self, key: str) -> "LazyLowerCaseSubFactory":
        lower_key = key.lower()
        return LazyLowerCaseSubFactory(
            self._factory, self._keys + (lower_key,), self._patype[lower_key].type
        )



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Definitions of different operations on a DataFrame."""

from typing import TYPE_CHECKING, Iterable, List

from tqdm import tqdm

from graviti.file import File, FileBase
from graviti.openapi import add_data, delete_data, update_data, update_schema
from graviti.operation.common import get_schema
from graviti.portex import record
from graviti.utility import submit_multithread_tasks

if TYPE_CHECKING:
    from graviti.dataframe import DataFrame
    from graviti.manager import ObjectPermissionManager

_MAX_BATCH_SIZE = 2048
_MAX_ITEMS = 60000


class DataFrameOperation:
    """This class defines the basic method of the operation on a DataFrame."""

    def get_file_count(self) -> int:  # pylint: disable=no-self-use
        """Get the file amount to be uploaded.

        Returns:
            The file amount to be uploaded.

        """
        return 0

    def get_data_count(self) -> int:  # pylint: disable=no-self-use
        """Get the data amount to be uploaded.

        Returns:
            The data amount to be uploaded.

        """
        return 0

    def do(  # pylint: disable=invalid-name
        self,
        access_key: str,
        url: str,
        owner: str,
        dataset: str,
        *,
        draft_number: int,
        sheet: str,
        jobs: int,
        data_pbar: tqdm,
        file_pbar: tqdm,
        object_permission_manager: "ObjectPermissionManager",
    ) -> None:
        """Execute the OpenAPI create sheet.

        Arguments:
            access_key: User's access key.
            url: The URL of the graviti website.
            owner: The owner of the dataset.
            dataset: Name of the dataset, unique for a user.
            draft_number: The draft number.
            sheet: The sheet name.
            jobs: The number of the max workers in multi-thread operation.
            data_pbar: The process bar for uploading structured data.
            file_pbar: The process bar for uploading binary files.
            object_permission_manager: The object permission manager of the dataset.

        Raises:
            NotImplementedError: The method of the base class should not be called.

        """
        raise NotImplementedError


class DataOperation(DataFrameOperation):  # pylint: disable=abstract-method
    """This class defines the basic method of the data operation on a DataFrame."""

    _files: List[FileBase]

    def __init__(self, data: "DataFrame") -> None:
        self._data = data

    def _get_max_batch_size(self) -> int:
        return min(
            _MAX_BATCH_SIZE,
            _MAX_ITEMS // self._data.schema._get_column_count(),  # pylint: disable=protected-access
        )

    def get_file_count(self) -> int:
        """Get the file amount to be uploaded.

        Returns:
            The file amount to be uploaded.

        """
        return len(list(self._data._generate_file()))  # pylint: disable=protected-access

    def get_data_count(self) -> int:
        """Get the data amount to be uploaded.

        Returns:
            The data amount to be uploaded.

        """
        return len(self._data)


class AddData(DataOperation):
    """This class defines the operation that add data to a DataFrame.

    Arguments:
        data: The data to be added.

    """

    def do(  # pylint: disable=invalid-name
        self,
        access_key: str,
        url: str,
        owner: str,
        dataset: str,
        *,
        draft_number: int,
        sheet: str,
        jobs: int,
        data_pbar: tqdm,
        file_pbar: tqdm,
        object_permission_manager: "ObjectPermissionManager",
    ) -> None:
        """Execute the OpenAPI add data.

        Arguments:
            access_key: User's access key.
            url: The URL of the graviti website.
            owner: The owner of the dataset.
            dataset: Name of the dataset, unique for a user.
            draft_number: The draft number.
            sheet: The sheet name.
            jobs: The number of the max workers in multi-thread operation.
            data_pbar: The process bar for uploading structured data.
            file_pbar: The process bar for uploading binary files.
            object_permission_manager: The object permission manager of the dataset.

        """
        batch_size = self._get_max_batch_size()
        df = self._data

        for i in range(0, len(df), batch_size):
            batch = df.iloc[i : i + batch_size]

            # pylint: disable=protected-access
            _upload_files(
                filter(
                    lambda f: isinstance(f, File), batch._generate_file()  # type: ignore[arg-type]
                ),
                object_permission_manager,
                jobs=jobs,
                pbar=file_pbar,
            )

            add_data(
                access_key,
                url,
                owner,
                dataset,
                draft_number=draft_number,
                sheet=sheet,
                data=batch.to_pylist(_to_backend=True),
            )
            data_pbar.update(len(batch))


class UpdateSchema(DataFrameOperation):
    """This class defines the operation that update the schema of a DataFrame.

    Arguments:
        schema: New portex schema after updated.

    """

    def __init__(self, schema: record) -> None:
        self.schema = schema

    def do(  # pylint: disable=invalid-name, unused-argument
        self,
        access_key: str,
        url: str,
        owner: str,
        dataset: str,
        *,
        draft_number: int,
        sheet: str,
        jobs: int,
        data_pbar: tqdm,
        file_pbar: tqdm,
        object_permission_manager: "ObjectPermissionManager",
    ) -> None:
        """Execute the OpenAPI update schema.

        Arguments:
            access_key: User's access key.
            url: The URL of the graviti website.
            owner: The owner of the dataset.
            dataset: Name of the dataset, unique for a user.
            draft_number: The draft number.
            sheet: The sheet name.
            jobs: The number of the max workers in multi-thread operation.
            data_pbar: The process bar for uploading structured data.
            file_pbar: The process bar for uploading binary files.
            object_permission_manager: The object permission manager of the dataset.

        """
        portex_schema, avro_schema, arrow_schema = get_schema(self.schema)

        update_schema(
            access_key,
            url,
            owner,
            dataset,
            draft_number=draft_number,
            sheet=sheet,
            _schema=portex_schema,
            _avro_schema=avro_schema,
            _arrow_schema=arrow_schema,
        )


class UpdateData(DataOperation):
    """This class defines the operation that updates the data of a DataFrame.

    Arguments:
        data: The data for updating.

    """

    def do(  # pylint: disable=invalid-name
        self,
        access_key: str,
        url: str,
        owner: str,
        dataset: str,
        *,
        draft_number: int,
        sheet: str,
        jobs: int,
        data_pbar: tqdm,
        file_pbar: tqdm,
        object_permission_manager: "ObjectPermissionManager",
    ) -> None:
        """Execute the OpenAPI add data.

        Arguments:
            access_key: User's access key.
            url: The URL of the graviti website.
            owner: The owner of the dataset.
            dataset: Name of the dataset, unique for a user.
            draft_number: The draft number.
            sheet: The sheet name.
            jobs: The number of the max workers in multi-thread operation.
            data_pbar: The process bar for uploading structured data.
            file_pbar: The process bar for uploading binary files.
            object_permission_manager: The object permission manager of the dataset.

        """
        batch_size = self._get_max_batch_size()
        df = self._data

        for i in range(0, len(df), batch_size):
            batch = df.iloc[i : i + batch_size]

            # pylint: disable=protected-access
            _upload_files(
                filter(
                    lambda f: isinstance(f, File), batch._generate_file()  # type: ignore[arg-type]
                ),
                object_permission_manager,
                jobs=jobs,
                pbar=file_pbar,
            )

            update_data(
                access_key,
                url,
                owner,
                dataset,
                draft_number=draft_number,
                sheet=sheet,
                data=batch.to_pylist(_to_backend=True),
            )
            data_pbar.update(len(batch))


class DeleteData(DataFrameOperation):
    """This class defines the operation that delete the data of a DataFrame.

    Arguments:
        record_keys: The record keys of the data to be deleted.

    """

    def __init__(self, record_keys: List[str]) -> None:
        self.record_keys = record_keys

    def do(  # pylint: disable=invalid-name, unused-argument
        self,
        access_key: str,
        url: str,
        owner: str,
        dataset: str,
        *,
        draft_number: int,
        sheet: str,
        jobs: int,
        data_pbar: tqdm,
        file_pbar: tqdm,
        object_permission_manager: "ObjectPermissionManager",
    ) -> None:
        """Execute the OpenAPI delete data.

        Arguments:
            access_key: User's access key.
            url: The URL of the graviti website.
            owner: The owner of the dataset.
            dataset: Name of the dataset, unique for a user.
            draft_number: The draft number.
            sheet: The sheet name.
            jobs: The number of the max workers in multi-thread operation.
            data_pbar: The process bar for uploading structured data.
            file_pbar: The process bar for uploading binary files.
            object_permission_manager: The object permission manager of the dataset.

        """
        delete_data(
            access_key,
            url,
            owner,
            dataset,
            draft_number=draft_number,
            sheet=sheet,
            record_keys=self.record_keys,
        )


def _upload_files(
    files: Iterable[File],
    object_permission_manager: "ObjectPermissionManager",
    pbar: tqdm,
    _allow_retry: bool = True,
    jobs: int = 8,
) -> None:
    submit_multithread_tasks(
        lambda file: _upload_file(file, object_permission_manager, pbar, _allow_retry),
        files,
        jobs=jobs,
    )


def _upload_file(
    file: File,
    object_permission_manager: "ObjectPermissionManager",
    pbar: tqdm,
    _allow_retry: bool = True,
) -> None:
    post_key = f"{object_permission_manager.prefix}{file.get_checksum()}"
    object_permission_manager.put_object(post_key, file.path, _allow_retry)
    file._post_key = post_key  # pylint: disable=protected-access
    pbar.update()



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""The implementation of the search related array."""

from typing import Any, Callable, ClassVar, Dict, Type, TypeVar

import pyarrow as pa

import graviti.portex as pt
from graviti.dataframe.sql.container import _E, ArrayContainer, SearchContainerRegister
from graviti.dataframe.sql.scalar import (
    NUMERICAL_PRIORITIES,
    BooleanScalar,
    DateScalar,
    EnumScalar,
    NumberScalar,
    RowSeries,
    StringScalar,
    TimedeltaScalar,
    TimeScalar,
    TimestampScalar,
)

_LOM = TypeVar("_LOM", bound="LogicalOperatorsMixin")
_EOM = TypeVar("_EOM", bound="EqualOperatorsMixin")
_COM = TypeVar("_COM", bound="ComparisonOperatorsMixin")
_AOM = TypeVar("_AOM", bound="ArithmeticOperatorsMixin")
_NA = TypeVar("_NA", bound="NumberArray")
_EA = TypeVar("_EA", bound="EnumArray")
_TAB = TypeVar("_TAB", bound="TemporalArrayBase")


class LogicalOperatorsMixin(ArrayContainer):
    """A mixin for dynamically implementing logical operators."""

    _LOCICAL_OPERATORS: ClassVar[Dict[str, str]] = {
        "__and__": "and",
        "__or__": "or",
    }

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        for meth, opt in cls._LOCICAL_OPERATORS.items():
            setattr(cls, meth, cls._get_logical_operator(opt))

    @classmethod
    def _get_logical_operator(cls: Type[_LOM], opt: str) -> Callable[[_LOM, Any], "BooleanArray"]:
        def func(self: _LOM, other: Any) -> "BooleanArray":
            other_expr = other.expr if isinstance(other, ArrayContainer) else other
            expr = {f"${opt}": [self.expr, other_expr]}

            return BooleanArray(expr, pt.boolean(), self.upper_expr)

        return func


class EqualOperatorsMixin(ArrayContainer):
    """A mixin for dynamically implementing euqal operators."""

    _EQUAL_OPERATORS: ClassVar[Dict[str, str]] = {
        "__eq__": "eq",
        "__ne__": "ne",
    }

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        for meth, opt in cls._EQUAL_OPERATORS.items():
            setattr(cls, meth, cls._get_equal_operator(opt))

    @classmethod
    def _get_equal_operator(cls: Type[_EOM], opt: str) -> Callable[[_EOM, Any], "BooleanArray"]:
        def func(self: _EOM, other: Any) -> "BooleanArray":
            other_expr = other.expr if isinstance(other, ArrayContainer) else other
            expr = {f"${opt}": [self.expr, other_expr]}

            return BooleanArray(expr, pt.boolean(), self.upper_expr)

        return func


class ComparisonOperatorsMixin(ArrayContainer):
    """A mixin for dynamically implementing comparison operators."""

    _COMPARISON_OPERATORS: ClassVar[Dict[str, str]] = {
        "__gt__": "gt",
        "__ge__": "gte",
        "__lt__": "lt",
        "__le__": "lte",
    }

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        for meth, opt in cls._COMPARISON_OPERATORS.items():
            setattr(cls, meth, cls._get_comparison_operator(opt))

    @classmethod
    def _get_comparison_operator(
        cls: Type[_COM], opt: str
    ) -> Callable[[_COM, Any], "BooleanArray"]:
        def func(self: _COM, other: Any) -> "BooleanArray":
            if isinstance(other, ArrayContainer):
                if not isinstance(other, type(self)):
                    raise TypeError(
                        f"Invalid '{opt}' operation between {self.schema} and {other.schema}"
                    )
                other_expr = other.expr
            else:
                other_expr = other

            expr = {f"${opt}": [self.expr, other_expr]}
            return BooleanArray(expr, pt.boolean(), self.upper_expr)

        return func


class ArithmeticOperatorsMixin(ArrayContainer):
    """A mixin for dynamically implementing arithmetic operators."""

    _ARITHMETIC_OPERATORS: ClassVar[Dict[str, str]] = {
        "__div__": "div",
        "__mod__": "mod",
        "__pow__": "pow",
        "__sub__": "sub",
        "__mul__": "mult",
        "__add__": "add",
    }

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        for meth, opt in cls._ARITHMETIC_OPERATORS.items():
            setattr(cls, meth, cls._get_arithmetic_operator(opt))

    @classmethod
    def _get_arithmetic_operator(cls: Type[_AOM], opt: str) -> Callable[[_AOM, Any], Any]:
        raise NotImplementedError


class Array(ArrayContainer):
    """One-dimensional array for portex builtin type array."""

    prefix = "$."

    def query(self, func: Callable[[Any], Any]) -> "Array":
        """Query the data of an ArraySeries with a lambda function.

        Arguments:
            func: The query function.

        Returns:
            The ArraySeries with the query expression.

        """
        item_container = self.schema.search_container.item_container
        expr = {
            "$filter": [
                self.upper_expr,
                func(item_container.from_upper(self.expr, self.schema)).expr,
            ]
        }

        return Array(expr, self.schema, self.upper_expr)

    def any(self) -> BooleanScalar:
        """Whether any element is True.

        Returns:
            The BooleanSeries with the any expression.

        """
        expr = {"$any_match": [self.upper_expr, self.expr]}
        return BooleanScalar(expr)

    def all(self) -> BooleanScalar:
        """Whether all elements are True.

        Returns:
            The BooleanSeries with the all expression.

        """
        expr = {"$all_match": [self.upper_expr, self.expr]}
        return BooleanScalar(expr)


@SearchContainerRegister(pt.boolean)
class BooleanArray(Array, LogicalOperatorsMixin, EqualOperatorsMixin):
    """One-dimensional array for portex builtin type array with the boolean items."""

    item_container = BooleanScalar


@SearchContainerRegister(pt.string)
class StringArray(Array, LogicalOperatorsMixin, EqualOperatorsMixin):
    """One-dimensional array for portex builtin type array with the string and enum items."""

    item_container = StringScalar


@SearchContainerRegister(pt.enum)
class EnumArray(Array, EqualOperatorsMixin):
    """One-dimensional array for portex builtin type array with the string and enum items."""

    item_container = EnumScalar

    @classmethod
    def _get_equal_operator(cls: Type[_EA], opt: str) -> Callable[[_EA, Any], "BooleanArray"]:
        def func(self: _EA, other: Any) -> "BooleanArray":
            if isinstance(other, ArrayContainer):
                other_expr: Any = other.expr
            else:
                enum: pt.enum = self.schema.to_builtin()  # type: ignore[assignment]
                other_expr = enum.values.value_to_index[other]

            expr = {f"${opt}": [self.expr, other_expr]}
            return BooleanArray(expr, pt.boolean(), self.upper_expr)

        return func


class TemporalArrayBase(Array, EqualOperatorsMixin, ComparisonOperatorsMixin):
    """One-dimensional array for portex builtin temporal types."""

    def _time_to_int(self, value: Any) -> int:
        return pa.scalar(value, self.schema.to_pyarrow()).value  # type: ignore[no-any-return]

    @classmethod
    def _get_equal_operator(cls: Type[_TAB], opt: str) -> Callable[[_TAB, Any], "BooleanArray"]:
        def func(self: _TAB, other: Any) -> "BooleanArray":
            if isinstance(other, ArrayContainer):
                other_expr: Any = other.expr
            else:
                other_expr = self._time_to_int(other)  # pylint: disable=protected-access

            expr = {f"${opt}": [self.expr, other_expr]}
            return BooleanArray(expr, pt.boolean(), self.upper_expr)

        return func

    @classmethod
    def _get_comparison_operator(
        cls: Type[_TAB], opt: str
    ) -> Callable[[_TAB, Any], "BooleanArray"]:
        def func(self: _TAB, other: Any) -> "BooleanArray":
            if isinstance(other, ArrayContainer):
                if not isinstance(other, type(self)):
                    raise TypeError(
                        f"Invalid '{opt}' operation between {self.schema} and {other.schema}"
                    )
                other_expr: Any = other.expr
            else:
                other_expr = self._time_to_int(other)  # pylint: disable=protected-access

            expr = {f"${opt}": [self.expr, other_expr]}
            return BooleanArray(expr, pt.boolean(), self.upper_expr)

        return func


@SearchContainerRegister(pt.date)
class DateArray(TemporalArrayBase):
    """One-dimensional array for portex builtin date type."""

    item_container = DateScalar

    def _time_to_int(self, value: Any) -> int:
        scalar = pa.scalar(value, self.schema.to_pyarrow())
        return scalar.cast(pa.int32()).as_py()  # type: ignore[no-any-return]


@SearchContainerRegister(pt.time)
class TimeArray(TemporalArrayBase):
    """One-dimensional array for portex builtin time type."""

    item_container = TimeScalar


@SearchContainerRegister(pt.timestamp)
class TimestampArray(TemporalArrayBase):
    """One-dimensional array for portex builtin timestamp type."""

    item_container = TimestampScalar


@SearchContainerRegister(pt.timedelta)
class TimedeltaArray(TemporalArrayBase):
    """One-dimensional array for portex builtin timedelta type."""

    item_container = TimedeltaScalar


@SearchContainerRegister(
    pt.float32,
    pt.float64,
    pt.int32,
    pt.int64,
)
class NumberArray(Array, ComparisonOperatorsMixin, ArithmeticOperatorsMixin):
    """One-dimensional array for portex builtin type array with the numerical items."""

    item_container = NumberScalar

    @classmethod
    def _get_arithmetic_operator(cls: Type[_NA], opt: str) -> Callable[[_NA, Any], _NA]:
        def func(self: _NA, other: Any) -> _NA:
            if isinstance(other, ArrayContainer):
                if not isinstance(other, type(self)):
                    raise TypeError(
                        f"Invalid '{opt}' operation between {self.schema} and {other.schema}"
                    )
                other_expr = other.expr
                schema = (
                    self.schema
                    if NUMERICAL_PRIORITIES[self.schema.__class__]
                    > NUMERICAL_PRIORITIES[other.schema.__class__]
                    else other.schema
                )
            else:
                other_expr = other
                schema = self.schema

            expr = {f"${opt}": [self.expr, other_expr]}
            return cls(expr, schema, self.upper_expr)

        return func

    def _check_upper_expr(self, opt: str) -> None:
        if self.expr != self.prefix:
            raise TypeError(f"{opt} operation only support for numerical array.")

    def size(self) -> NumberScalar:
        """Get the length of array series.

        Returns:
            The NumberScalar with the size expression.

        """
        self._check_upper_expr("size")
        return NumberScalar({"$size": [self.upper_expr]}, self.schema)

    def max(self) -> "NumberScalar":
        """Get the max value of array series.

        Returns:
            The NumberScalar with the max expression.

        """
        self._check_upper_expr("max")
        return NumberScalar({"$max": [self.upper_expr]}, self.schema)

    def min(self) -> NumberScalar:
        """Get the min value of array series.

        Returns:
            The NumberScalar with the min expression.

        """
        self._check_upper_expr("min")
        return NumberScalar({"$min": [self.upper_expr]}, self.schema)

    def sum(self) -> NumberScalar:
        """Get the sum of array series.

        Returns:
            The NumberScalar with the sum expression.

        """
        self._check_upper_expr("sum")
        return NumberScalar({"$sum": [self.upper_expr]}, self.schema)


@SearchContainerRegister(pt.record)
class DataFrame(ArrayContainer):
    """The Two-dimensional array for the search."""

    prefix = "$"
    item_container = RowSeries
    schema: pt.PortexRecordBase

    def __getitem__(self, key: str) -> ArrayContainer:
        field = self.schema[key]
        return field.search_container(f"{self.expr}.{key}", field, self.upper_expr)

    def query(self, func: Callable[[Any], Any]) -> "DataFrame":
        """Query the data of an ArraySeries with a lambda function.

        Arguments:
            func: The query function.

        Returns:
            The DataFrame with the query expression.

        """
        expr = {"$filter": [self.upper_expr, func(RowSeries(self.schema)).expr]}
        return DataFrame(expr, self.schema, self.upper_expr)


class ArrayDistributor(ArrayContainer):
    """A distributor to instance DataFrame, ArrayScalar by different array items."""

    @classmethod
    def from_upper(cls, expr: _E, schema: pt.PortexType) -> "ArrayContainer":
        """Instantiate a Search object from the upper level.

        Arguments:
            expr: The expression of the search.
            schema: The schema of the series.

        Returns:
            The loaded object.

        """
        items: pt.PortexType = schema.to_builtin().items  # type: ignore[attr-defined]
        return items.search_container.from_upper(expr, items)


@SearchContainerRegister(pt.array)
class ArraySeries(ArrayContainer):
    """The One-dimensional array for the search."""

    item_container = ArrayDistributor



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""The implementation of iteration tools."""

from itertools import islice
from typing import Iterable, Iterator, Tuple, TypeVar

_T = TypeVar("_T")


def chunked(iterable: Iterable[_T], n: int) -> Iterator[Tuple[_T, ...]]:
    """Break an iterable instance into tuples of length n.

    Arguments:
        iterable: The input iterable instance which needs to be breaked into tuples of length n.
        n: The length of each yielded tuples.

    Yields:
        The tuples of length n.

    Examples:
        >>> list(chunked(range(9), 3))
        [(0, 1, 2), (3, 4, 5), (6, 7, 8)]

        The last yielded tuple may have fewer than n items if the length of the input iterable
        instance is not divisible by n:

        >>> list(chunked(range(10), 3))
        [(0, 1, 2), (3, 4, 5), (6, 7, 8), (9,)]

    """
    iterator = iter(iterable)

    while True:
        chunk = tuple(islice(iterator, n))
        if not chunk:
            return

        yield chunk



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#
"""Utility module."""

from graviti.utility.attr import AttrDict
from graviti.utility.collections import (
    FrozenNameOrderedDict,
    NameOrderedDict,
    UserMapping,
    UserMutableMapping,
    UserMutableSequence,
    UserSequence,
)
from graviti.utility.common import (
    ModuleMocker,
    convert_datetime_to_gmt,
    convert_iso_to_datetime,
    locked,
    shorten,
    urlnorm,
)
from graviti.utility.engine import Mode, engine
from graviti.utility.itertools import chunked
from graviti.utility.repr import INDENT, MAX_REPR_ROWS, ReprMixin, ReprType
from graviti.utility.requests import UserResponse, config, get_session, submit_multithread_tasks
from graviti.utility.typing import NestedDict, PathLike, check_type

__all__ = [
    "AttrDict",
    "FrozenNameOrderedDict",
    "INDENT",
    "MAX_REPR_ROWS",
    "Mode",
    "ModuleMocker",
    "NameOrderedDict",
    "NestedDict",
    "PathLike",
    "ReprMixin",
    "ReprType",
    "UserMapping",
    "UserMutableMapping",
    "UserMutableSequence",
    "UserResponse",
    "UserSequence",
    "check_type",
    "chunked",
    "config",
    "convert_datetime_to_gmt",
    "convert_iso_to_datetime",
    "engine",
    "get_session",
    "locked",
    "shorten",
    "submit_multithread_tasks",
    "urlnorm",
]



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Operation module."""

from graviti.operation.frame import (
    AddData,
    DataFrameOperation,
    DeleteData,
    UpdateData,
    UpdateSchema,
)
from graviti.operation.sheet import CreateSheet, DeleteSheet, SheetOperation

__all__ = [
    "AddData",
    "CreateSheet",
    "DataFrameOperation",
    "DeleteData",
    "DeleteSheet",
    "SheetOperation",
    "UpdateData",
    "UpdateSchema",
]



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""The implementation of the Graviti Series."""

from itertools import chain, zip_longest
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union, overload

from graviti.dataframe.row.indexing import RowSeriesILocIndexer, RowSeriesLocIndexer
from graviti.utility import MAX_REPR_ROWS


class Series:
    """One-dimensional array.

    Arguments:
        data: The data that needs to be stored in Series. Could be ndarray or Iterable.
        schema: Data type to force. If None, will be inferred from ``data``.
        index: Index of the ``data``.

    Examples:
        Constructing Series from a list.

        >>> d = {"filename": "a.jpg", "attributes": {"color": "red", "pose": "frontal"}}
        >>> series = Series(data=d)
        >>> series
        filename           a.jpg
        attributes color     red
                    pose frontal

    """

    _indices_data: Dict[str, Any]
    _indices: List[str]

    def __init__(
        self,
        data: Optional[Dict[str, Any]] = None,
        schema: Any = None,
        index: Optional[List[str]] = None,
    ) -> None:
        if data is None:
            data = {}
        if schema is not None:
            # TODO: missing schema processing
            pass
        if index is not None:
            # TODO: missing index processing
            pass

        self._indices_data, self._indices = {}, []
        for key, value in data.items():
            if isinstance(value, dict):
                value = Series(value)
            self._indices_data[key] = value
            self._indices.append(key)

    def __repr__(self) -> str:
        flatten_header, flatten_data = self._flatten()
        header = self._get_repr_header(flatten_header)
        body = [
            item._repr_folding() if hasattr(item, "_repr_folding") else str(item)
            for item in flatten_data
        ]
        column_widths = [len(max(row, key=len)) for row in chain(header, [body])]
        lines = [
            "".join(f"{item:<{column_widths[index]+2}}" for index, item in enumerate(line))
            for line in zip_longest(*header, body)
        ]
        if self.__len__() > MAX_REPR_ROWS:
            lines.append(f"...({self.__len__()})")
        return "\n".join(lines)

    # @overload
    # def __getitem__(self, key: Union[slice, Tuple[str]]) -> "Series":
    #    ...

    @overload
    def __getitem__(self, key: str) -> Any:
        ...

    @overload
    def __getitem__(self, key: Iterable[str]) -> "Series":
        ...

    def __getitem__(self, key: Union[str, Iterable[str]]) -> Any:
        if isinstance(key, str):
            return self._indices_data[key]

        new_data = {name: self._indices_data[name] for name in key}
        return Series(new_data)

    def __setitem__(self, key: str, value: Any) -> None:
        pass

    def __len__(self) -> int:
        return self._indices.__len__()

    @staticmethod
    def _get_repr_header(flatten_header: List[Tuple[str, ...]]) -> List[List[str]]:
        lines: List[List[str]] = []
        for names in zip_longest(*flatten_header, fillvalue=""):
            line = []
            pre_name = None
            upper_line = lines[-1][1:] if lines else []
            for name, upper_name in zip_longest(names, upper_line, fillvalue=""):
                if name == pre_name and upper_name == "":
                    line.append("")
                else:
                    line.append(name)
                pre_name = name
            lines.append(line)
        return lines

    @classmethod
    def _construct(cls, indices_data: Dict[str, Any]) -> "Series":
        obj: Series = object.__new__(cls)
        # pylint: disable=protected-access
        obj._indices_data = indices_data
        obj._indices = list(indices_data.keys())
        return obj

    def _flatten(self) -> Tuple[List[Tuple[str, ...]], List[Any]]:
        header: List[Tuple[str, ...]] = []
        data: List[Any] = []
        for key, value in self._indices_data.items():
            if isinstance(value, Series):
                nested_header, nested_data = value._flatten()  # pylint: disable=protected-access
                header.extend((key, *sub_column) for sub_column in nested_header)
                data.extend(nested_data)
            else:
                data.append(value)
                header.append((key,))
        return header, data

    # @overload
    # def _getitem_by_location(self, key: slice) -> "Series":
    #    ...

    @overload
    def _getitem_by_location(self, key: int) -> Union["Series", Any]:
        ...

    @overload
    def _getitem_by_location(self, key: Iterable[int]) -> "Series":
        ...

    def _getitem_by_location(self, key: Union[int, Iterable[int]]) -> Any:
        if isinstance(key, int):
            return self._indices_data[self._indices[key]]

        indices_data = {
            self._indices[index]: self._indices_data[self._indices[index]] for index in key
        }
        return self._construct(indices_data)

    @property
    def iloc(self) -> RowSeriesILocIndexer:
        """Purely integer-location based indexing for selection by position.

        Allowed inputs are:

        - An integer, e.g. ``5``.
        - A list or array of integers, e.g. ``[4, 3, 0]``.
        - A slice object with ints, e.g. ``1:7``.
        - A boolean array of the same length as the axis being sliced.

        Returns:
            The instance of the ILocIndexer.

        Examples:
            >>> series = Series([1, 2, 3])
            >>> series.loc[0]
            1
            >>> df.loc[[0]]
            0    1
            dtype: int64

        """
        return RowSeriesILocIndexer(self)

    @property
    def loc(self) -> RowSeriesLocIndexer:
        """Access a group of rows and columns by indexes or a boolean array.

        Allowed inputs are:

        - A single index, e.g. ``5``.
        - A list or array of indexes, e.g. ``[4, 3, 0]``.
        - A slice object with indexes, e.g. ``1:7``.
        - A boolean array of the same length as the axis being sliced.

        Returns:
            The instance of the LocIndexer.

        Examples:
            >>> series = Series([1, 2, 3])
            >>> series.loc[0]
            1
            >>> df.loc[[0]]
            0    1
            dtype: int64

        """
        return RowSeriesLocIndexer(self)



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""The implementation of the Graviti Series."""


from itertools import islice
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)

import pyarrow as pa

import graviti.portex as pt
from graviti.dataframe.column.indexing import ColumnSeriesILocIndexer, ColumnSeriesLocIndexer
from graviti.dataframe.container import RECORD_KEY, Container
from graviti.file import FileBase
from graviti.operation import UpdateData
from graviti.paging import (
    LazyFactoryBase,
    MappedPagingList,
    PagingList,
    PagingListBase,
    PyArrowPagingList,
)
from graviti.portex.enum import EnumValueList, EnumValueType
from graviti.utility import MAX_REPR_ROWS, ModuleMocker

try:
    import pandas as pd
except ModuleNotFoundError:
    pd = ModuleMocker("No module named 'pandas'")

if TYPE_CHECKING:
    import pandas

    from graviti.dataframe.frame import DataFrame
    from graviti.manager.permission import ObjectPermissionManager

_S = TypeVar("_S", bound="Series")
_P = TypeVar("_P", bound="PyarrowSeries")
_A = TypeVar("_A", bound="ArraySeries")
_F = TypeVar("_F", bound="FileSeries")
_E = TypeVar("_E", bound="EnumSeries")

_OPM = Optional["ObjectPermissionManager"]


class Series(Container):
    """One-dimensional array.

    Arguments:
        data: The data that needs to be stored in Series. Could be ndarray or Iterable.
        schema: Data type to force. If None, will be inferred from ``data``.

    Examples:
        Constructing Series from a list.

        >>> d = [1,2,3,4]
        >>> series = Series(data=d)
        >>> series
        0 1
        1 2
        2 3
        3 4

    """

    schema: pt.PortexType
    _data: PagingListBase[Any]

    def __new__(
        cls: Type[_S],
        data: Union[Iterable[Any], _S],
        schema: Optional[pt.PortexType] = None,
    ) -> Any:
        """One-dimensional data with schema.

        Arguments:
            data: The data that needs to be stored in column series.
            schema: The schema of the column series. If None, will be inferred from `data`.

        Raises:
            ValueError: When both given schema and Series data.
            ValueError: When using DataFrame as data to construct a Series.

        Returns:
            The created Series object.

        """
        if isinstance(data, cls):
            if schema is not None:
                raise ValueError("Only support the schema when data is not Series")

            return data.copy()

        if data.__class__.__name__ == "DataFrame":
            raise ValueError("Do not support Constructing a Series through a DataFrame")

        if schema is None:
            array = pa.array(data)
            schema = pt.PortexType.from_pyarrow(array.type)
        elif issubclass(schema.container, FileSeries):
            return FileSeries._from_iterable(data, schema)
        else:
            array = cls._pylist_to_pyarrow(data, schema)

        return schema.container._from_pyarrow(array, schema)

    def __repr__(self) -> str:
        indices = list(self._get_repr_indices())
        indice_width = len(str(max(indices, default=1)))

        body = []
        body_item_width = 0
        for i in indices:
            item = self.loc[i]
            name = item._repr_folding() if hasattr(item, "_repr_folding") else str(item)
            body.append(name)
            body_item_width = max(len(name), body_item_width)

        lines = []
        for indice, value in zip(indices, body):
            lines.append(f"{indice:<{indice_width+2}}{value:<{body_item_width+2}}")
        if self.__len__() > MAX_REPR_ROWS:
            lines.append(f"...({self.__len__()})")
        return "\n".join(lines)

    def __len__(self) -> int:
        return self._data.__len__()

    @overload
    def __getitem__(self: _S, key: slice) -> _S:
        ...

    @overload
    def __getitem__(self, key: int) -> Any:
        ...

    # @overload
    # def __getitem__(self, key: Iterable[int]) -> "Series":
    #     ...

    def __getitem__(self: _S, key: Union[int, slice]) -> Union[Any, _S]:
        if isinstance(key, int):
            return self._get_item_by_location(key)

        return self._get_slice_by_location(key, self.schema.copy())

    @overload
    def __setitem__(self, key: slice, value: Union[Iterable[Any], _S]) -> None:
        ...

    @overload
    def __setitem__(self, key: int, value: Any) -> None:
        ...

    def __setitem__(
        self,
        key: Union[int, slice],
        value: Union[Iterable[Any], _S, Any],
    ) -> None:
        if isinstance(key, int):
            series = self._from_iterable([value], self.schema)
            key = slice(key, key + 1)
        elif not isinstance(value, self.__class__):
            series = self._from_pyarrow(self._pylist_to_pyarrow(value, self.schema), self.schema)
        elif self.schema.to_pyarrow().equals(value.schema.to_pyarrow()):
            series = value
        else:
            raise TypeError("The schema of the given Series is mismatched")

        self._set_slice(key, series)
        if self._root is not None and self._root.operations is not None:
            value = series.copy()
            root = self._root
            name = self._name
            df = root._create(pt.record({name[0]: series.schema.copy()}), None, name[1:])
            df._columns = {name[0]: series}
            df[RECORD_KEY] = root._record_key[key]  # type: ignore[index, assignment]

            root.operations.append(UpdateData(df))  # type: ignore[union-attr]

    def __delitem__(self, key: Union[int, slice]) -> None:
        if self._root is not None:
            raise TypeError(
                "'__delitem__' is not supported for the Series which is a member of a DataFrame"
            )

        self._del_item_by_location(key)

    def __iter__(self) -> Iterator[Any]:
        return self._data.__iter__()

    @staticmethod
    def _pylist_to_pyarrow(values: Iterable[Any], schema: pt.PortexType) -> pa.StructArray:
        if isinstance(values, Iterator):
            values = list(values)
        processor, need_process = Series._process_array(values, schema)

        if not need_process:
            return pa.array(values, schema.to_pyarrow())

        return pa.array(processor(values), schema.to_pyarrow())

    @staticmethod
    def _process_array(
        values: Iterable[Any], schema: pt.PortexType
    ) -> Tuple[Callable[[Any], Any], Optional[bool]]:
        if not values:
            return lambda x: x, None

        for value in values:
            if value is None:
                continue

            processor, need_process = Series._get_process(value, schema)
            if need_process is None:
                continue

            return lambda x: list(map(processor, x)), need_process

        return lambda x: x, False

    @staticmethod
    def _get_process(
        value: Any, schema: pt.PortexType
    ) -> Tuple[Callable[[Any], Any], Optional[bool]]:
        container = schema.container
        if value.__class__.__name__ == "DataFrame":
            return lambda x: x.to_pylist(_to_backend=True), True

        if container == ArraySeries:
            return Series._process_array(
                value, schema.to_builtin().items  # type: ignore[attr-defined]
            )

        if container == FileSeries:
            return Series._process_file(value)

        if container == EnumSeries:
            value_to_index = schema.to_builtin().values.value_to_index  # type: ignore[attr-defined]
            return lambda x: value_to_index[x], True

        return lambda x: x, False

    @staticmethod
    def _process_file(values: Iterable[Any]) -> Tuple[Callable[[Any], Any], bool]:
        if isinstance(values, FileBase):
            return lambda x: x.to_pyobj(), True

        return lambda x: x, False

    @classmethod
    def _from_factory(
        cls: Type[_S],
        factory: LazyFactoryBase,
        schema: pt.PortexType,
        root: Optional["DataFrame"] = None,
        name: Tuple[str, ...] = (),
        *,
        object_permission_manager: _OPM = None,
    ) -> _S:
        obj = cls._create(schema, root, name)
        obj._refresh_data_from_factory(factory, object_permission_manager)
        return obj

    @classmethod
    def _from_iterable(
        cls: Type[_S],
        array: Iterable[Any],
        schema: pt.PortexType,
    ) -> _S:
        return cls(array, schema)

    def _repr_folding(self) -> str:
        return f"{self.__class__.__name__}({self.__len__()})"

    def _get_repr_indices(self) -> Iterable[int]:
        return islice(range(len(self)), MAX_REPR_ROWS)

    def _get_item_by_location(self, key: int) -> Any:
        return self._data.get_item(key)

    def _get_slice_by_location(
        self: _S,
        key: slice,
        schema: pt.PortexType,
        root: Optional["DataFrame"] = None,
        name: Tuple[str, ...] = (),
    ) -> _S:
        obj = self._create(schema, root, name)
        obj._data = self._data.get_slice(key)  # pylint: disable=protected-access

        return obj

    def _set_slice(self: _S, key: slice, value: _S) -> None:
        self._data.set_slice(key, value._data)  # pylint: disable=protected-access

    def _del_item_by_location(self, key: Union[int, slice]) -> None:
        self._data.__delitem__(key)

    def _refresh_data_from_factory(self, factory: LazyFactoryBase, _: _OPM) -> None:
        self._data = factory.create_pyarrow_list()

    def _extend(self: _S, values: _S) -> None:
        """Extend Series to itself row by row.

        Arguments:
            values: A series that needs to be extended.

        Examples:
            >>> s1 = Series.from_pyarrow(pa.array([1,2,3]))
            >>> s1
            0  1
            1  2
            2  3
            >>> s2 = Series.from_pyarrow(pa.array([4,5,6]))
            >>> s2
            0  4
            1  5
            2  6
            >>> s1.extend(s2)
            >>> s1
            0  1
            1  2
            2  3
            3  1
            4  2
            5  3

        """
        self._data.extend(values._data)  # pylint: disable=protected-access

    def _copy(
        self: _S,
        schema: pt.PortexType,
        root: Optional["DataFrame"] = None,
        name: Tuple[str, ...] = (),
    ) -> _S:
        obj = self._create(schema, root, name)
        obj._data = self._data.copy()  # pylint: disable=protected-access

        return obj

    def _to_pandas_series(self) -> "pandas.Series":
        return self.to_pandas()

    @classmethod
    def from_pyarrow(cls, array: pa.Array) -> "Series":
        """Instantiate a Series backed by an pyarrow array.

        Arguments:
            array: The input pyarrow array.

        Raises:
            TypeError: When the input pyarrow type is not supported.

        Returns:
            The loaded :class:`~graviti.dataframe.column.Series` instance.

        """
        pyarrow_type = array.type

        # pylint: disable=protected-access
        if pa.types.is_dictionary(pyarrow_type):
            schema: pt.PortexType = pt.enum(array.dictionary.to_pylist())

            return EnumSeries._from_pyarrow(array.indices, schema)

        schema = pt.PortexType.from_pyarrow(array.type)
        container = schema.container
        if not issubclass(container, Series):
            raise TypeError(
                f"pyarrow type '{pyarrow_type}' is not supported converting to a graviti Series"
            )

        return container._from_pyarrow(array, schema)

    @classmethod
    def from_pandas(cls, series: "pandas.Series") -> "Series":
        """Convert a pandas Series to a graviti Series.

        Arguments:
            series: The input pandas Series.

        Returns:
            The converted graviti Series.

        """
        array = pa.Array.from_pandas(series)

        return cls.from_pyarrow(array)

    @property
    def iloc(self) -> ColumnSeriesILocIndexer:
        """Purely integer-location based indexing for selection by position.

        Allowed inputs are:

        - An integer, e.g. ``5``.
        - A list or array of integers, e.g. ``[4, 3, 0]``.
        - A slice object with ints, e.g. ``1:7``.
        - A boolean array of the same length as the axis being sliced.

        Returns:
            The instance of the ILocIndexer.

        Examples:
            >>> series = Series([1, 2, 3])
            >>> series.loc[0]
            1
            >>> df.loc[[0]]
            0    1
            dtype: int64

        """
        return ColumnSeriesILocIndexer(self)

    @property
    def loc(self) -> ColumnSeriesLocIndexer:
        """Access a group of rows and columns by indexes or a boolean array.

        Allowed inputs are:

        - A single index, e.g. ``5``.
        - A list or array of indexes, e.g. ``[4, 3, 0]``.
        - A slice object with indexes, e.g. ``1:7``.
        - A boolean array of the same length as the axis being sliced.

        Returns:
            The instance of the LocIndexer.

        Examples:
            >>> series = Series([1, 2, 3])
            >>> series.loc[0]
            1
            >>> df.loc[[0]]
            0    1
            dtype: int64

        """
        return ColumnSeriesLocIndexer(self)

    def to_pylist(self, *, _to_backend: bool = False) -> List[Any]:
        """Convert the container to a python list.

        Raises:
            NotImplementedError: The method of the base class should not be called.

        """
        raise NotImplementedError

    def to_pandas(self) -> "pandas.Series":
        """Convert the graviti Container to a pandas Series or DataFrame.

        Raises:
            NotImplementedError: The method of the base class should not be called.

        """
        raise NotImplementedError


class PyarrowSeries(Series):
    """Pyarrow based one-dimensional array."""

    _data: PyArrowPagingList[Any]

    @classmethod
    def _from_pyarrow(
        cls: Type[_P],
        array: pa.Array,
        schema: pt.PortexType,
        root: Optional["DataFrame"] = None,
        name: Tuple[str, ...] = (),
        *,
        object_permission_manager: _OPM = None,  # pylint: disable=unused-argument
    ) -> _P:
        obj = cls._create(schema, root, name)
        obj._data = PyArrowPagingList.from_pyarrow(array)

        return obj

    def __iter__(self) -> Iterator[Any]:
        return (i.as_py() for i in self._data)

    def _get_item_by_location(self, key: int) -> Any:
        return self._data.get_item(key).as_py()

    def to_pylist(self, *, _to_backend: bool = False) -> List[Any]:
        """Convert the Series to a python list.

        Returns:
            The python list representing the Series.

        """
        return self._data.to_pyarrow().to_pylist()  # type: ignore[no-any-return]

    def to_pandas(self) -> "pandas.Series":
        """Convert the graviti Series to a pandas Series.

        Returns:
            The converted pandas Series.

        """
        return self._data.to_pyarrow().to_pandas()


@pt.ContainerRegister(
    pt.boolean,
    pt.float32,
    pt.float64,
    pt.int32,
    pt.int64,
)
class NumberSeries(PyarrowSeries):
    """One-dimensional array for portex builtin number type."""


@pt.ContainerRegister(pt.string)
class StringSeries(PyarrowSeries):
    """One-dimensional array for portex builtin string type."""


@pt.ContainerRegister(pt.binary)
class BinarySeries(PyarrowSeries):
    """One-dimensional array for portex builtin binary type."""


@pt.ContainerRegister(pt.array)
class ArraySeries(Series):
    """One-dimensional array for portex builtin type array."""

    _data: MappedPagingList[Any]
    _item_schema: pt.PortexType
    _object_permission_manager: _OPM = None

    @classmethod
    def _from_pyarrow(
        cls: Type[_A],
        array: pa.ListArray,
        schema: pt.PortexType,
        root: Optional["DataFrame"] = None,
        name: Tuple[str, ...] = (),
        *,
        object_permission_manager: _OPM = None,
    ) -> _A:
        builtin_schema: pt.array = schema.to_builtin()  # type: ignore[assignment]
        _item_schema = builtin_schema.items
        _item_creator = _item_schema.container._from_pyarrow  # pylint: disable=protected-access

        obj = cls._create(schema, root, name)
        obj._data = MappedPagingList.from_array(
            array,
            lambda scalar: _item_creator(
                scalar.values, _item_schema, object_permission_manager=object_permission_manager
            ),
        )

        obj._item_schema = _item_schema
        obj._object_permission_manager = object_permission_manager

        return obj

    @classmethod
    def _from_iterable(
        cls: Type[_A],
        array: Iterable[Any],
        schema: pt.PortexType,
    ) -> _A:
        obj: _A = object.__new__(cls)
        items_schema = schema.to_builtin().items  # type: ignore[attr-defined]
        items_container = items_schema.container
        array = (
            items_container._from_iterable(value, items_schema)  # pylint: disable=protected-access
            for value in array
        )
        obj._data = MappedPagingList(array)
        obj.schema = schema
        obj._item_schema = items_schema
        return obj

    def _get_slice_by_location(
        self: _A,
        key: slice,
        schema: pt.PortexType,
        root: Optional["DataFrame"] = None,
        name: Tuple[str, ...] = (),
    ) -> _A:
        obj = super()._get_slice_by_location(key, schema, root, name)
        # pylint: disable=protected-access
        obj._object_permission_manager = self._object_permission_manager
        return obj._copy(schema, root, name)

    def _refresh_data_from_factory(
        self, factory: LazyFactoryBase, object_permission_manager: _OPM
    ) -> None:
        builtin_schema: pt.array = self.schema.to_builtin()  # type: ignore[assignment]
        _item_schema = builtin_schema.items
        _item_creator = _item_schema.container._from_pyarrow  # pylint: disable=protected-access

        self._data = factory.create_mapped_list(
            lambda scalar: _item_creator(
                scalar.values, _item_schema, object_permission_manager=object_permission_manager
            )
        )
        self._item_schema = _item_schema  # pylint: disable=protected-access
        self._object_permission_manager = object_permission_manager

    def _extract_paging_list(self: _A, values: _A) -> MappedPagingList[Any]:
        # pylint: disable=protected-access
        _item_schema = self._item_schema
        if values._item_schema is _item_schema and values is not self:
            return values._data

        _item_creator = _item_schema.container._from_pyarrow
        return values._data.copy(
            lambda df: df._copy(_item_schema),
            lambda scalar: _item_creator(
                scalar.values,
                _item_schema,
                object_permission_manager=values._object_permission_manager,
            ),
        )

    def _set_item_by_slice(self: _A, key: slice, value: _A) -> None:
        self._data.set_slice(key, self._extract_paging_list(value))

    def _extend(self: _A, values: _A) -> None:
        self._data.extend(self._extract_paging_list(values))

    def _copy(
        self: _A,
        schema: pt.PortexType,
        root: Optional["DataFrame"] = None,
        name: Tuple[str, ...] = (),
    ) -> _A:
        obj = self._create(schema, root, name)

        builtin_schema: pt.array = schema.to_builtin()  # type: ignore[assignment]
        _item_schema = builtin_schema.items
        _item_creator = _item_schema.container._from_pyarrow  # pylint: disable=protected-access
        _object_permission_manager = self._object_permission_manager

        # pylint: disable=protected-access
        obj._item_schema = _item_schema
        obj._data = self._data.copy(
            lambda df: df._copy(_item_schema),
            lambda scalar: _item_creator(
                scalar.values, _item_schema, object_permission_manager=_object_permission_manager
            ),
        )
        obj._object_permission_manager = _object_permission_manager

        return obj

    def to_pylist(self, *, _to_backend: bool = False) -> List[Any]:
        """Convert the Series to a python list.

        Returns:
            The python list representing the Series.

        """
        return [item.to_pylist(_to_backend=_to_backend) for item in self._data]

    def to_pandas(self) -> "pandas.Series":
        """Convert the graviti Series to a pandas Series.

        Returns:
            The converted pandas Series.

        """
        return pd.Series(item.to_pylist() for item in self._data)


@pt.ExternalContainerRegister(
    pt.STANDARD_URL,
    "main",
    "file.File",
    "file.Audio",
    "file.Image",
    "file.PointCloud",
    "file.PointCloudBin",
    "label.Mask",
)
class FileSeries(Series):
    """One-dimensional array for file."""

    _data: PagingList[Any]

    @classmethod
    def _from_iterable(
        cls: Type[_F],
        array: Iterable[FileBase],
        schema: pt.PortexType,
    ) -> _F:
        obj: _F = object.__new__(cls)
        obj._data = PagingList(array)
        obj.schema = schema
        return obj

    @classmethod
    def _from_pyarrow(
        cls: Type[_F],
        array: pa.StructArray,
        schema: pt.PortexType,
        root: Optional["DataFrame"] = None,
        name: Tuple[str, ...] = (),
        *,
        object_permission_manager: _OPM = None,
    ) -> _F:
        if object_permission_manager is None:
            raise ValueError(
                "The object permission manager is needed to create FileSeries from pyarrow"
            )

        obj = cls._create(schema, root, name)
        file_type: FileBase = schema.element  # type: ignore[assignment]
        # pylint: disable=protected-access
        obj._data = PagingList(
            file_type._from_pyarrow(item, object_permission_manager) for item in array
        )

        return obj

    def _refresh_data_from_factory(
        self, factory: LazyFactoryBase, object_permission_manager: _OPM
    ) -> None:
        file_type = self.schema.element
        self._data = factory.create_list(
            lambda scalar: file_type(
                **scalar.as_py(), object_permission_manager=object_permission_manager
            )
        )

    def to_pylist(self, *, _to_backend: bool = False) -> List[Any]:
        """Convert the BinaryFileSeries to python list.

        Returns:
            The python list.

        """
        if _to_backend:
            return [file._to_post_data() for file in self._data]  # pylint: disable=protected-access

        return list(self._data)

    def to_pandas(self) -> "pandas.Series":
        """Convert the graviti Series to a pandas Series.

        Returns:
            The converted pandas Series.

        """
        return pd.Series(self._data)


@pt.ContainerRegister(pt.enum)
class EnumSeries(PyarrowSeries):
    """One-dimensional array for portex builtin type enum."""

    _index_to_value: Dict[Optional[int], EnumValueType]

    def _get_item_by_location(self, key: int) -> Any:
        return self._index_to_value[self._data[key].as_py()]

    def _get_slice_by_location(
        self: _E,
        key: slice,
        schema: pt.PortexType,
        root: Optional["DataFrame"] = None,
        name: Tuple[str, ...] = (),
    ) -> _E:
        obj = super()._get_slice_by_location(key, schema, root, name)
        enum_values = self.schema.to_builtin().values  # type: ignore[attr-defined]
        obj._index_to_value = enum_values.index_to_value  # pylint: disable=protected-access

        return obj

    def _copy(
        self,
        schema: pt.PortexType,
        root: Optional["DataFrame"] = None,
        name: Tuple[str, ...] = (),
    ) -> "EnumSeries":
        obj = super()._copy(schema, root, name)
        enum_values = self.schema.to_builtin().values  # type: ignore[attr-defined]
        obj._index_to_value = enum_values.index_to_value  # pylint: disable=protected-access

        return obj

    @classmethod
    def _from_pyarrow(
        cls: Type[_E],
        array: pa.IntegerArray,
        schema: pt.PortexType,
        root: Optional["DataFrame"] = None,
        name: Tuple[str, ...] = (),
        *,
        object_permission_manager: _OPM = None,  # pylint: disable=unused-argument
    ) -> _E:
        obj = cls._create(schema, root, name)
        obj._data = PyArrowPagingList.from_pyarrow(array)

        enum_values = schema.to_builtin().values  # type: ignore[attr-defined]
        obj._index_to_value = enum_values.index_to_value

        return obj

    def _refresh_data_from_factory(self, factory: LazyFactoryBase, _: _OPM) -> None:
        self._data = factory.create_pyarrow_list()
        enum_values = self.schema.to_builtin().values  # type: ignore[attr-defined]
        self._index_to_value = enum_values.index_to_value

    def to_pylist(self, *, _to_backend: bool = False) -> List[Any]:
        """Convert the Series to a python list.

        Returns:
            The python list representing the Series.

        """
        if _to_backend:
            return super().to_pylist()

        _index_to_value = self._index_to_value
        return [_index_to_value[i.as_py()] for i in self._data]

    def to_pandas(self) -> "pandas.Series":
        """Convert the graviti EnumSeries to a pandas Categorical Series.

        Returns:
            The converted pandas Categorical Series.

        Raises:
            TypeError: When the enum values in 'dict' format.

        """
        enum_values = self.schema.to_builtin().values  # type: ignore[attr-defined]
        if not isinstance(enum_values, EnumValueList):
            raise TypeError(
                "The enum values in 'dict' format is not supported to converted to pandas"
            )

        dictionary = enum_values.to_pyobj()

        array = self._data.to_pyarrow().combine_chunks()
        dictionary_array = pa.DictionaryArray.from_arrays(array, dictionary)
        return dictionary_array.to_pandas()


@pt.ContainerRegister(pt.date, pt.time, pt.timestamp, pt.timedelta)
class TimeSeries(PyarrowSeries):
    """One-dimensional array for portex builtin temporal type."""

    def to_pylist(self, *, _to_backend: bool = False) -> List[Any]:
        """Convert the Series to a python list.

        Returns:
            The python list representing the Series.

        """
        if _to_backend:
            array = self._data.to_pyarrow()
            bit_width = array.type.bit_width
            target_type = pa.int32() if bit_width == 32 else pa.int64()

            return array.cast(target_type).to_pylist()  # type: ignore[no-any-return]

        return super().to_pylist()



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#
"""Manager module."""

from graviti.manager.branch import BranchManager
from graviti.manager.commit import Commit, CommitManager
from graviti.manager.dataset import Dataset, DatasetManager
from graviti.manager.draft import Draft, DraftManager
from graviti.manager.permission import (
    AZUREObjectPermissionManager,
    ObjectPermissionManager,
    OSSObjectPermissionManager,
    S3ObjectPermissionManager,
)
from graviti.manager.tag import TagManager

__all__ = [
    "AZUREObjectPermissionManager",
    "BranchManager",
    "Commit",
    "CommitManager",
    "Dataset",
    "DatasetManager",
    "Draft",
    "DraftManager",
    "OSSObjectPermissionManager",
    "ObjectPermissionManager",
    "S3ObjectPermissionManager",
    "TagManager",
]



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""The implementation of the Graviti DataFrame."""

# pylint: disable=too-many-lines

from itertools import chain, islice, zip_longest
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import pyarrow as pa

import graviti.portex as pt
from graviti.dataframe.column.series import ArraySeries, EnumSeries, FileSeries, NumberSeries
from graviti.dataframe.column.series import Series as ColumnSeries
from graviti.dataframe.container import RECORD_KEY, Container
from graviti.dataframe.indexing import DataFrameILocIndexer, DataFrameLocIndexer
from graviti.dataframe.row.series import Series as RowSeries
from graviti.dataframe.sql import RowSeries as SqlRowSeries
from graviti.file import FileBase
from graviti.operation import AddData, DataFrameOperation, DeleteData, UpdateData, UpdateSchema
from graviti.paging import LazyFactoryBase
from graviti.utility import MAX_REPR_ROWS, Mode, ModuleMocker, engine

try:
    import pandas as pd
except ModuleNotFoundError:
    pd = ModuleMocker("No module named 'pandas'")


if TYPE_CHECKING:
    import pandas

    from graviti.manager.permission import ObjectPermissionManager


_T = TypeVar("_T", bound="DataFrame")
_C = TypeVar("_C", bound="Container")

_OPM = Optional["ObjectPermissionManager"]

APPLY_KEY = "apply_result"


def _to_simple_pyarrow(schema: pt.PortexType) -> pa.DataType:
    if not isinstance(schema, pt.PortexRecordBase):
        return schema.to_pyarrow()

    fields = []
    for name, portex_type in schema.items():
        builtin_type = portex_type.to_builtin()
        if (
            isinstance(builtin_type, pt.array)
            or issubclass(portex_type.container, FileSeries)
            or issubclass(portex_type.container, EnumSeries)
        ):
            continue

        if isinstance(portex_type, pt.PortexRecordBase):
            sub_fields = _to_simple_pyarrow(portex_type)
            if sub_fields:
                fields.append(pa.field(name, sub_fields))
        else:
            fields.append(pa.field(name, _to_simple_pyarrow(portex_type)))

    return pa.struct(fields)


@pt.ContainerRegister(pt.record)
class DataFrame(Container):
    """Two-dimensional, size-mutable, potentially heterogeneous tabular data.

    Arguments:
        data: The data that needs to be stored in DataFrame.
        schema: The schema of the DataFrame. If None, will be inferred from `data`.
        columns: Column labels to use for resulting frame when data does not have them,
            defaulting to RangeIndex(0, 1, 2, ..., n). If data contains column labels,
            will perform column selection instead.

    Examples:
        Constructing DataFrame from list.

        >>> df = DataFrame(
        ...     [
        ...         {"filename": "a.jpg", "box2ds": {"x": 1, "y": 1}},
        ...         {"filename": "b.jpg", "box2ds": {"x": 2, "y": 2}},
        ...         {"filename": "c.jpg", "box2ds": {"x": 3, "y": 3}},
        ...     ]
        ... )
        >>> df
            filename box2ds
                     x      y
        0   a.jpg    1      1
        1   b.jpg    2      2
        2   c.jpg    3      3

    """

    _columns: Dict[str, Container]
    _record_key: Optional[NumberSeries] = None

    schema: pt.PortexRecordBase
    operations: Optional[List[DataFrameOperation]] = None
    searcher: Optional[Callable[[Dict[str, Any], pt.PortexRecordBase], "DataFrame"]] = None
    criteria: Optional[Dict[str, Any]] = None

    def __new__(
        cls: Type[_T],
        data: Union[List[Dict[str, Any]], _T, None] = None,
        schema: Optional[pt.PortexRecordBase] = None,
    ) -> _T:
        """Two-dimensional, size-mutable, potentially heterogeneous tabular data.

        Arguments:
            data: The data that needs to be stored in DataFrame.
            schema: The schema of the DataFrame. If None, will be inferred from `data`.

        Raises:
            ValueError: When both given schema and DataFrame data.
            ValueError: When the given data is not list.

        Returns:
            The created :class:`~graviti.dataframe.DataFrame` object.

        """
        if data is None:
            data = []

        if isinstance(data, cls):
            if schema is not None:
                raise ValueError("Only support the schema when data is not DataFrame")

            return data.copy()

        if schema is None:
            return cls.from_pyarrow(pa.array(data))

        if isinstance(data, list):
            return cls._from_pyarrow_and_pydict(
                *cls._pylist_to_pyarrow_and_pydict(data, schema), schema
            )

        raise ValueError("DataFrame only supports creating from list object now")

    def __repr__(self) -> str:
        flatten_header, flatten_data = self._flatten()
        header = self._get_repr_header(flatten_header)
        body = self._get_repr_body(flatten_data)
        column_widths = [len(max(column, key=len)) for column in zip_longest(*header, *body)]
        lines = [
            "".join(f"{item:<{column_widths[index]+2}}" for index, item in enumerate(line))
            for line in chain(header, body)
        ]
        if self.__len__() > MAX_REPR_ROWS:
            lines.append(f"...({self.__len__()})")
        return "\n".join(lines)

    def __len__(self) -> int:
        return next(self._columns.values().__iter__()).__len__()

    # @overload
    # def __getitem__(self, key: str) -> Union[ColumnSeriesBase, "DataFrame"]:  # type: ignore[misc]
    #     # https://github.com/python/mypy/issues/5090
    #     ...

    # @overload
    # def __getitem__(self, key: Iterable[str]) -> "DataFrame":
    #     ...

    def __getitem__(self, key: str) -> Container:
        if isinstance(key, str):
            return self._columns[key]

        # if isinstance(key, Iterable):
        #     new_columns = {name: self._columns[name] for name in key}
        #     schema = pt.record({name: column.schema for name, column in new_columns.items()})
        #     return self._construct(new_columns, schema, self._record_key)

        raise KeyError(key)

    def __setitem__(self, key: str, value: Union[Iterable[Any], _C]) -> None:
        root = self if self._root is None else self._root
        if root.operations is None:
            self._setitem(key, value)
            return

        if key in self.schema:
            raise NotImplementedError("Replacing a column in a sheet is not supported yet")

        self._setitem(key, value)
        # TODO: Use self[[key]] to replace DataFrame._construct
        _value = self[key].copy()

        df = self._create(pt.record({key: _value.schema}), None, self._name)
        df._columns = {key: _value}
        df[RECORD_KEY] = root._record_key  # type: ignore[index, assignment]

        assert isinstance(self.schema, pt.record)
        root.operations.extend((UpdateSchema(self.schema), UpdateData(df)))

    @staticmethod
    def _get_repr_header(flatten_header: List[Tuple[str, ...]]) -> List[List[str]]:
        lines: List[List[str]] = []
        for names in zip_longest(*flatten_header, fillvalue=""):
            line = [""]
            pre_name = None
            upper_line = lines[-1][1:] if lines else [""]
            for name, upper_name in zip_longest(names, upper_line, fillvalue=""):
                line.append("" if name == pre_name and upper_name == "" else name)
                pre_name = name
            lines.append(line)
        return lines

    @staticmethod
    def _get_process(
        value: Any, schema: pt.PortexType
    ) -> Tuple[Callable[[Any], Any], Optional[bool]]:
        container = schema.container
        if container == DataFrame:
            return DataFrame._process_record(value, schema)  # type: ignore[arg-type]
        if container == ArraySeries:
            return DataFrame._process_array(
                value, schema.to_builtin().items  # type: ignore[attr-defined]
            )
        if container == FileSeries:
            return DataFrame._process_file(value)
        if container == EnumSeries:
            values_to_indices: Dict[Union[int, float, str, bool, None], Optional[int]] = {
                k: v for v, k in enumerate(schema.to_builtin().values)  # type: ignore[attr-defined]
            }
            if None not in values_to_indices:
                values_to_indices[None] = None
            return lambda x: values_to_indices[x], True

        return lambda x: x, False

    @staticmethod
    def _process_array(
        values: Iterable[Any], schema: pt.PortexType
    ) -> Tuple[Callable[[Any], Any], Optional[bool]]:
        if isinstance(values, DataFrame):
            return lambda x: x.to_pylist(_to_backend=True), True

        if not values:
            return lambda x: x, None

        for value in values:
            if value is None:
                continue

            processor, need_process = DataFrame._get_process(value, schema)
            if need_process is None:
                continue

            return lambda x: list(map(processor, x)), need_process
        return lambda x: x, False

    @staticmethod
    def _process_file(values: Iterable[Any]) -> Tuple[Callable[[Any], Any], bool]:
        if isinstance(values, FileBase):
            return lambda x: x.to_pyobj(), True

        return lambda x: x, False

    @staticmethod
    def _process_record(
        values: Dict[str, Any], schema: pt.PortexRecordBase
    ) -> Tuple[Callable[[Any], Any], Optional[bool]]:
        processors: Dict[str, Callable[[Any], Any]] = {}
        need_process: Optional[bool] = False
        sub_need_process: Optional[bool] = False
        for name, field in schema.items():
            item = values[name]
            if item is None:
                processors[name] = lambda x: x
            else:
                processors[name], sub_need_process = DataFrame._get_process(item, field)
                need_process = need_process or sub_need_process

        return lambda x: {k: processors[k](v) for k, v in x.items()}, need_process

    @staticmethod
    def _pylist_to_pyarrow(values: Iterable[Any], schema: pt.PortexRecordBase) -> pa.StructArray:

        processor, need_process = DataFrame._process_array(values, schema)

        if not need_process:
            return pa.array(values, schema.to_pyarrow())

        return pa.array(processor(values), schema.to_pyarrow())

    @staticmethod
    def _pylist_to_pyarrow_and_pydict(
        values: Iterable[Any], schema: pt.PortexRecordBase
    ) -> Tuple[pa.StructArray, Dict[str, Any]]:
        pyarrow_schema = _to_simple_pyarrow(schema)
        pyarrow_array = pa.array(values, pyarrow_schema)
        pydict = DataFrame._get_pydict(values, schema)

        return pyarrow_array, pydict

    @staticmethod
    def _get_pydict(values: Iterable[Any], schema: pt.PortexRecordBase) -> Dict[str, Any]:
        pydict: Dict[str, Any] = {}
        for key, portex_type in schema.items():
            container = portex_type.container
            if (
                issubclass(container, ArraySeries)
                or issubclass(container, FileSeries)
                or issubclass(container, EnumSeries)
            ):
                pydict[key] = container._from_iterable(  # pylint: disable=protected-access
                    (value[key] for value in values), portex_type
                )
            elif isinstance(portex_type, pt.PortexRecordBase):
                sub_pydict = DataFrame._get_pydict([value[key] for value in values], portex_type)
                if sub_pydict:
                    pydict[key] = sub_pydict
        return pydict

    @classmethod
    def _from_pyarrow(  # type: ignore[override]
        cls: Type[_T],
        array: pa.StructArray,
        schema: pt.PortexRecordBase,
        root: Optional["DataFrame"] = None,
        name: Tuple[str, ...] = (),
        *,
        object_permission_manager: _OPM = None,
    ) -> _T:
        obj = cls._create(schema, root, name)

        obj._columns = {
            key: value.container._from_pyarrow(  # pylint: disable=protected-access
                array.field(key),
                schema=value,
                root=obj if root is None else root,
                name=(key,) + name,
                object_permission_manager=object_permission_manager,
            )
            for key, value in schema.items()
        }
        return obj

    @classmethod
    def _from_factory(  # type: ignore[override]
        cls: Type[_T],
        factory: LazyFactoryBase,
        schema: pt.PortexRecordBase,
        root: Optional["DataFrame"] = None,
        name: Tuple[str, ...] = (),
        *,
        object_permission_manager: _OPM = None,
    ) -> _T:
        """Create DataFrame with paging lists.

        Arguments:
            factory: The LazyFactory instance for creating the PagingList.
            schema: The schema of the DataFrame.
            root: The root of the created DataFrame.
            name: The name of the created DataFrame.
            object_permission_manager: The ObjectPermissionManager instance for reading files.

        Returns:
            The loaded :class:`~graviti.dataframe.DataFrame` object.

        """
        obj = cls._create(schema, root, name)

        obj._columns = {
            key: value.container._from_factory(  # pylint: disable=protected-access
                factory[key],
                schema=value,
                root=obj if root is None else root,
                name=(key,) + name,
                object_permission_manager=object_permission_manager,
            )
            for key, value in schema.items()
        }

        if RECORD_KEY in factory:
            obj._record_key = NumberSeries._from_factory(  # pylint: disable=protected-access
                factory[RECORD_KEY], pt.string(nullable=True)
            )

        return obj

    @classmethod
    def _from_iterable(  # type: ignore[override]
        cls: Type[_T],
        array: Union[List[Dict[str, Any]], _T, None],
        schema: pt.PortexRecordBase,
    ) -> _T:
        if isinstance(array, cls):
            return array._copy(schema)  # pylint: disable=protected-access
        return cls(array, schema)

    @classmethod
    def _from_pyarrow_and_pydict(  # pylint: disable=too-many-arguments
        cls: Type[_T],
        array: pa.StructArray,
        pydict: Dict[str, Any],
        schema: pt.PortexRecordBase,
        root: Optional["DataFrame"] = None,
        name: Tuple[str, ...] = (),
    ) -> _T:
        obj = cls._create(schema, root, name)

        columns: Dict[str, Any] = {}
        for key, value in schema.items():
            sub_pydict = pydict.get(key, {})
            if isinstance(sub_pydict, (FileSeries, ArraySeries, EnumSeries)):
                sub_pydict._root = (  # pylint: disable=protected-access
                    root if root is not None else obj
                )
                columns[key] = sub_pydict
                continue

            if isinstance(value, pt.PortexRecordBase):
                sub_array = (
                    array.field(key) if array.type.get_field_index(key) != -1 else pa.struct([])
                )
                columns[key] = cls._from_pyarrow_and_pydict(
                    sub_array,
                    sub_pydict,
                    schema=value,
                    root=obj if root is None else root,
                    name=(key,) + name,
                )
            else:
                columns[key] = value.container._from_pyarrow(  # pylint: disable=protected-access
                    array.field(key),
                    schema=value,
                    root=obj if root is None else root,
                    name=(key,) + name,
                )
        obj._columns = columns

        return obj

    def _refresh_data_from_factory(
        self, factory: LazyFactoryBase, object_permission_manager: _OPM
    ) -> None:
        for key, column in self._columns.items():
            column._refresh_data_from_factory(  # pylint: disable=protected-access
                factory[key], object_permission_manager
            )

    def _get_item_by_location(self, key: int) -> RowSeries:
        indices_data = {
            name: self._columns[name]._get_item_by_location(key)  # pylint: disable=protected-access
            for name in self.schema
        }
        return RowSeries._construct(indices_data)  # pylint: disable=protected-access

    def _get_slice_by_location(  # type: ignore[override]
        self: _T,
        key: slice,
        schema: pt.PortexRecordBase,
        root: Optional["DataFrame"] = None,
        name: Tuple[str, ...] = (),
    ) -> _T:
        obj = self._create(schema, root, name)

        _root = obj if root is None else root
        _columns = self._columns
        columns = {}

        # pylint: disable=protected-access
        for column_name, value in schema.items():
            columns[column_name] = _columns[column_name]._get_slice_by_location(
                key,
                value,
                _root,  # type: ignore[arg-type]
                (column_name,) + name,
            )

        obj._columns = columns

        return obj

    def _set_slice_by_location(self, key: slice, value: "DataFrame") -> None:
        for name, column in self._columns.items():
            column.loc[key] = value[name]

    def _del_item_by_location(self, key: Union[int, slice]) -> None:
        for column in self._columns.values():
            column._del_item_by_location(key)  # pylint: disable=protected-access
        if self.operations is not None:
            record_key: NumberSeries = self._record_key  # type: ignore[assignment]
            if isinstance(key, int):
                record_keys = [record_key[key]]
            else:
                record_keys = list(record_key[key])
            self.operations.append(DeleteData(record_keys))

    def _setitem(self, key: str, value: Union[Iterable[Any], _C]) -> None:
        root = self if self._root is None else self._root
        if isinstance(value, Container):
            schema = value.schema.copy()
            column = value._copy(  # pylint: disable=protected-access
                schema, root, (key,) + self._name
            )
        else:
            # TODO: Need to support the case where iterable elements are subclass of Container.
            array = pa.array(value)
            schema = pt.PortexType.from_pyarrow(array.type)
            column = schema.container._from_pyarrow(  # pylint: disable=protected-access
                array, schema, root, (key,) + self._name
            )

        if len(column) != len(self):
            raise ValueError(
                f"Length of values ({len(column)}) does not match length of "
                f"self DataFrame ({len(self)})"
            )
        self.schema[key] = schema
        self._columns[key] = column

    def _flatten(self) -> Tuple[List[Tuple[str, ...]], List[ColumnSeries]]:
        header: List[Tuple[str, ...]] = []
        data: List[ColumnSeries] = []
        for key, value in self._columns.items():
            if isinstance(value, DataFrame):
                nested_header, nested_data = value._flatten()  # pylint: disable=protected-access
                header.extend((key, *sub_column) for sub_column in nested_header)
                data.extend(nested_data)
            elif isinstance(value, ColumnSeries):
                data.append(value)
                header.append((key,))
            else:
                raise TypeError(f"Unsupported column type: {value.__class__}")

        return header, data

    def _repr_folding(self) -> str:
        return f"{self.__class__.__name__}{self.shape}"

    def _get_repr_indices(self) -> Iterable[int]:
        return islice(range(len(self)), MAX_REPR_ROWS)

    def _get_repr_body(self, flatten_data: List[ColumnSeries]) -> List[List[str]]:
        lines = []
        for i in self._get_repr_indices():
            line: List[str] = [str(i)]
            for column in flatten_data:
                item = column[i]
                name = (
                    item._repr_folding()  # pylint: disable=protected-access
                    if isinstance(item, Container)
                    else str(item)
                )
                line.append(name)
            lines.append(line)
        return lines

    def _generate_file(self) -> Iterator[FileBase]:
        for extrator in _generate_file_extractor(self.schema):
            yield from extrator(self)

    def _extend(self, values: "DataFrame") -> None:
        for key, value in self._columns.items():
            value._extend(values[key])  # pylint: disable=protected-access

    def _to_pandas_series(self) -> "pandas.Series":
        return pd.Series(self.to_pylist())

    def _copy(  # type: ignore[override]
        self: _T,
        schema: pt.PortexRecordBase,
        root: Optional["DataFrame"] = None,
        name: Tuple[str, ...] = (),
    ) -> _T:
        obj = self._create(schema, root, name)

        _root = obj if root is None else root
        _columns = self._columns
        columns = {}

        # pylint: disable=protected-access
        for key, value in schema.items():
            column = _columns[key]._copy(
                value,
                _root,  # type: ignore[arg-type]
                (key,) + name,
            )
            columns[key] = column

        obj._columns = columns

        return obj

    @classmethod
    def from_pyarrow(cls: Type[_T], array: pa.StructArray) -> _T:
        """Create DataFrame with pyarrow struct array.

        Arguments:
            array: The input pyarrow struct array.

        Returns:
            The loaded :class:`~graviti.dataframe.DataFrame` instance.

        """
        schema = pt.record.from_pyarrow(array.type)
        return cls._from_pyarrow(array, schema)

    @classmethod
    def from_pandas(cls: Type[_T], df: "pandas.DataFrame") -> _T:
        """Create DataFrame with pandas DataFrame.

        Arguments:
            df: The input pandas DataFrame.

        Raises:
            NotImplementedError: When the column index of input DataFrame is MultiIndex.

        Returns:
            The loaded :class:`~graviti.dataframe.DataFrame` instance.

        """
        if isinstance(df.columns, pd.MultiIndex):
            raise NotImplementedError("pandas 'MultiIndex' is not supported now")

        record_batch = pa.RecordBatch.from_pandas(df)
        array = pa.StructArray.from_arrays(record_batch.columns, record_batch.schema.names)

        return cls.from_pyarrow(array)

    @property
    def iloc(self) -> DataFrameILocIndexer:
        """Purely integer-location based indexing for selection by position.

        Allowed inputs are:

        - An integer, e.g. ``5``.
        - A tuple, e.g. ``(5, "COLUMN_NAME")``

        Returns:
            The instance of the ILocIndexer.

        Examples:
            >>> df = DataFrame({"col1": [1, 2], "col2": [3, 4]})
            >>> df.iloc[0]
            col1    1
            col2    3
            Name: 0, dtype: int64
            >>> df.iloc[0, "col1"]
            1

        """
        return DataFrameILocIndexer(self)

    @property
    def loc(self) -> DataFrameLocIndexer:
        """Access the row by indexes.

        Allowed inputs are:

        - A single index, e.g. ``5``.
        - A tuple, e.g. ``(5, "COLUMN_NAME")``

        Returns:
            The instance of the LocIndexer.

        Examples:
            >>> df = DataFrame({"col1": [1, 2], "col2": [3, 4]})
            >>> df.loc[0]
            col1    1
            col2    3
            Name: 0, dtype: int64
            >>> df.loc[0, "col1"]
            1

        """
        return DataFrameLocIndexer(self)

    @property
    def shape(self) -> Tuple[int, int]:
        """Return a tuple representing the dimensionality of the DataFrame.

        Returns:
            Shape of the DataFrame.

        Examples:
            >>> df = DataFrame(
            ...     [
            ...         {"filename": "a.jpg", "box2ds": {"x": 1, "y": 1}},
            ...         {"filename": "b.jpg", "box2ds": {"x": 2, "y": 2}},
            ...         {"filename": "c.jpg", "box2ds": {"x": 3, "y": 3}},
            ...     ]
            ... )
            >>> df
                filename box2ds
                         x      y
            0   a.jpg    1      1
            1   b.jpg    2      2
            2   c.jpg    3      3
            >>> df.shape
            (3, 2)

        """
        return (self.__len__(), len(self.schema))

    @property
    def size(self) -> int:
        """Return an int representing the number of elements in this object.

        Returns:
            Size of the DataFrame.

        Examples:
            >>> df = DataFrame({"col1": [1, 2], "col2": [3, 4]})
            >>> df.size
            4

        """
        return self.__len__() * len(self.schema)

    def head(self: _T, n: int = 5) -> _T:
        """Return the first `n` rows.

        Arguments:
            n: Number of rows to select.

        Returns:
            The first `n` rows.

        Examples:
            >>> df = DataFrame(
            ...     [
            ...         {"animal": "alligator"},
            ...         {"animal": "bee"},
            ...         {"animal": "falcon"},
            ...         {"animal": "lion"},
            ...         {"animal": "monkey"},
            ...         {"animal": "parrot"},
            ...         {"animal": "shark"},
            ...         {"animal": "whale"},
            ...         {"animal": "zebra"},
            ...     ]
            ... )
            >>> df
                  animal
            0  alligator
            1        bee
            2     falcon
            3       lion
            4     monkey
            5     parrot
            6      shark
            7      whale
            8      zebra

            Viewing the first `n` lines (three in this case)

            >>> df.head(3)
                  animal
            0  alligator
            1        bee
            2     falcon

            For negative values of `n`

            >>> df.head(-3)
                  animal
            0  alligator
            1        bee
            2     falcon
            3       lion
            4     monkey
            5     parrot

        """
        return self._get_slice_by_location(slice(0, n), self.schema.copy())

    def tail(self: _T, n: int = 5) -> _T:
        """Return the last `n` rows.

        Arguments:
            n: Number of rows to select.

        Returns:
            The last `n` rows.

        Examples:
            >>> df = DataFrame(
            ...     [
            ...         {"animal": "alligator"},
            ...         {"animal": "bee"},
            ...         {"animal": "falcon"},
            ...         {"animal": "lion"},
            ...         {"animal": "monkey"},
            ...         {"animal": "parrot"},
            ...         {"animal": "shark"},
            ...         {"animal": "whale"},
            ...         {"animal": "zebra"},
            ...     ]
            ... )
            >>> df
                  animal
            0  alligator
            1        bee
            2     falcon
            3       lion
            4     monkey
            5     parrot
            6      shark
            7      whale
            8      zebra

            Viewing the last 5 lines

            >>> df.tail()
               animal
            0  monkey
            1  parrot
            2   shark
            3   whale
            4   zebra

            Viewing the last `n` lines (three in this case)

            >>> df.tail(3)
              animal
            0  shark
            1  whale
            2  zebra

        """
        return self._get_slice_by_location(slice(-n, None), self.schema.copy())

    # def sample(self, n: Optional[int] = None, axis: Optional[int] = None) -> "DataFrame":
    #     """Return a random sample of items from an axis of object.
    #
    #     Arguments:
    #         n: Number of items from axis to return.
    #         axis: {0 or `index`, 1 or `columns`, None}
    #             Axis to sample. Accepts axis number or name. Default is stat axis
    #             for given data type (0 for Series and DataFrames).
    #
    #     Return:
    #         A new object of same type as caller containing `n` items randomly
    #         sampled from the caller object.
    #
    #     """

    # def info(self) -> None:
    #     """Print a concise summary of a DataFrame."""

    def extend(self, values: Union[Iterable[Dict[str, Any]], "DataFrame"]) -> None:
        """Extend Sequence object or DataFrame to itself row by row.

        Arguments:
            values: A sequence object or DataFrame.

        Raises:
            TypeError: When the self is the member of another Dataframe.
            TypeError: When the given Dataframe mismatched with the self schema.

        Examples:
            >>> df = DataFrame([
            ...     {"filename": "a.jpg", "box2ds": {"x": 1, "y": 1}},
            ...     {"filename": "b.jpg", "box2ds": {"x": 2, "y": 2}},
            ... ])

            Extended by another list.

            >>> df.extend([{"filename": "c.jpg", "box2ds": {"x": 3, "y": 3}}])
            >>> df
                filename box2ds
                         x      y
            0   a.jpg    1      1
            1   b.jpg    2      2
            2   c.jpg    3      3

            Extended by another DataFrame.

            >>> df2 = DataFrame([{"filename": "d.jpg", "box2ds": {"x": 4 "y": 4}}])
            >>> df.extend(df2)
            >>> df
                filename box2ds
                         x      y
            0   a.jpg    1      1
            1   b.jpg    2      2
            2   d.jpg    4      4

        """
        if self._root is not None:
            raise TypeError(
                "'extend' is not supported for the DataFrame which is a member of another DataFrame"
            )
        if not isinstance(values, DataFrame):
            values = DataFrame._from_pyarrow_and_pydict(  # pylint: disable=protected-access
                *self._pylist_to_pyarrow_and_pydict(values, self.schema), self.schema
            )
        elif not self.schema.to_pyarrow().equals(values.schema.to_pyarrow()):
            raise TypeError("The schema of the given DataFrame is mismatched")
        elif self.operations is not None:
            values = values.copy()

        values_length = len(values)
        self._extend(values)
        if self.operations is not None:
            self.operations.append(AddData(values))

        if self._record_key is not None:
            self._record_key._data.extend_nulls(values_length)  # pylint: disable=protected-access

    def to_pylist(self, *, _to_backend: bool = False) -> List[Dict[str, Any]]:
        """Convert the DataFrame to a python list.

        Returns:
            The python list representing the DataFrame.

        """
        column_names = self.schema.keys()
        return [
            dict(zip(column_names, values))
            for values in zip(
                *(column.to_pylist(_to_backend=_to_backend) for column in self._columns.values())
            )
        ]

    def to_pandas(self) -> "pandas.DataFrame":
        """Convert the graviti DataFrame to a pandas DataFrame.

        Returns:
            The converted pandas DataFrame.

        """
        df = pd.DataFrame()

        for key in self.schema.keys():
            df[key] = self._columns[key]._to_pandas_series()  # pylint: disable=protected-access

        return df

    def query(self, func: Callable[[Any], Any]) -> "DataFrame":
        """Query the columns of a DataFrame with a lambda function.

        Arguments:
            func: The query function.

        Returns:
            The query result DataFrame.

        Raises:
            TypeError: When the DataFrame is not in a Commit.

        Examples:
            >>> df = DataFrame([
            ...     {"filename": "a.jpg", "box2ds": {"x": 1, "y": 1}},
            ...     {"filename": "b.jpg", "box2ds": {"x": 2, "y": 2}},
            ... ])
            >>> df.query(lambda x: x["filename"] == "a.jpg")
                filename box2ds
                         x      y
            0   a.jpg    1      1

        """
        if engine.mode != Mode.ONLINE:
            raise TypeError(
                "Only online mode is supported now, "
                "Please use 'engine.online' with block to enable the online mode."
            )
        if self.searcher is None:
            raise TypeError("'query' is not supported for the DataFrame not in a Commit")

        result = func(SqlRowSeries(self.schema))
        if self.criteria is None:
            criteria = {"where": result.expr}
        else:
            criteria = {
                "where": {
                    "$and": [
                        self.criteria["where"],  # pylint: disable=unsubscriptable-object
                        result.expr,
                    ]
                }
            }

        df = self.searcher(criteria, self.schema)  # pylint: disable=not-callable
        df.criteria = criteria
        df.searcher = self.searcher
        return df

    def apply(self, func: Callable[[Any], Any]) -> Container:
        """Apply a function to the DataFrame row by row.

        Arguments:
            func: Function to apply to each row.

        Returns:
            The apply result DataFrame or Series.

        Raises:
            TypeError: When the DataFrame is not in a Commit.

        Examples:
            >>> df = DataFrame([
            ...     {"filename": "a.jpg", "box2ds": {"x": 1, "y": 1}},
            ...     {"filename": "b.jpg", "box2ds": {"x": 2, "y": 2}},
            ... ])
            >>> df.apply(lambda x: x["box2ds"]["x"] + 1)
                filename box2ds
                         x      y
            0   a.jpg    2      1
            1   b.jpg    3      2

        """
        if engine.mode != Mode.ONLINE:
            raise TypeError(
                "Only online mode is supported now, "
                "Please use 'engine.online' with block to enable the online mode."
            )
        if self.searcher is None:
            raise TypeError("'apply' is not supported for the DataFrame not in a Commit")

        result = func(SqlRowSeries(self.schema))
        criteria = {} if self.criteria is None else self.criteria.copy()
        criteria["select"] = [{APPLY_KEY: result.expr}]
        df = self.searcher(  # pylint: disable=not-callable
            criteria, pt.record({APPLY_KEY: pt.array(result.schema)})
        )
        return df[APPLY_KEY]


def _generate_file_extractor(
    schema: pt.PortexRecordBase, base_extrator: Callable[[DataFrame], Any] = lambda df: df
) -> Iterator[Callable[[DataFrame], Iterable[FileBase]]]:
    for key, value in schema.items():

        def extrator(df: DataFrame, _key: str = key) -> Any:
            return base_extrator(df)[_key]

        container = value.container
        if container == FileSeries:
            yield extrator
        elif container == DataFrame:
            yield from _generate_file_extractor(value, extrator)  # type: ignore[arg-type]
        elif container == ArraySeries:
            yield from _generate_file_extrator_from_array(
                value, lambda df, _e=extrator: (_e(df),)  # type: ignore[misc]
            )


def _generate_file_extrator_from_record(
    schema: pt.PortexRecordBase, base_extrator: Callable[[DataFrame], Iterable[Any]]
) -> Iterator[Callable[[DataFrame], Iterator[FileBase]]]:
    for key, value in schema.items():
        yield from _assign_extrator(
            value,
            lambda df, _key=key: (item[_key] for item in base_extrator(df)),  # type: ignore[misc]
        )


def _generate_file_extrator_from_array(
    schema: pt.PortexType, base_extrator: Callable[[DataFrame], Iterable[Any]]
) -> Iterator[Callable[[DataFrame], Iterator[FileBase]]]:
    yield from _assign_extrator(
        schema.to_builtin().items,  # type: ignore[attr-defined]
        lambda df: chain.from_iterable(base_extrator(df)),
    )


def _assign_extrator(
    schema: pt.PortexType, extrator: Callable[[DataFrame], Iterable[Any]]
) -> Iterator[Callable[[DataFrame], Iterator[FileBase]]]:
    container = schema.container

    if container == FileSeries:
        yield lambda df: chain.from_iterable(extrator(df))
    elif container == DataFrame:
        yield from _generate_file_extrator_from_record(schema, extrator)  # type: ignore[arg-type]
    elif container == ArraySeries:
        yield from _generate_file_extrator_from_array(schema, extrator)
